from typing import Tuple

from dlt.common import logger
from dlt.extract.source import DltSource


def join_url(base_url: str, path: str) -> str:
    if not base_url.endswith("/"):
        base_url += "/"
    return base_url + path.lstrip("/")


def check_connection(
    source: DltSource,
    *resource_names: str,
) -> Tuple[bool, str]:
    try:
        list(source.with_resources(*resource_names).add_limit(1))
        return (True, "")
    except Exception as e:
        logger.error(f"Error checking connection: {e}")
        return (False, str(e))



from typing import Tuple, Union, Optional

from dlt.common.typing import TimedeltaSeconds

# Either a single timeout or tuple (connect,read) timeout
TRequestTimeout = Union[TimedeltaSeconds, Tuple[TimedeltaSeconds, TimedeltaSeconds]]



from __future__ import annotations

from typing import Any, Type, Union, Dict, TYPE_CHECKING

from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.destinations.impl.duckdb.factory import _set_duckdb_raw_capabilities
from dlt.destinations.impl.motherduck.configuration import (
    MotherDuckCredentials,
    MotherDuckClientConfiguration,
)

if TYPE_CHECKING:
    from duckdb import DuckDBPyConnection
    from dlt.destinations.impl.motherduck.motherduck import MotherDuckClient


class motherduck(Destination[MotherDuckClientConfiguration, "MotherDuckClient"]):
    spec = MotherDuckClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps = _set_duckdb_raw_capabilities(caps)
        caps.preferred_loader_file_format = "parquet"
        caps.max_query_length = 512 * 1024
        caps.max_parallel_load_jobs = 8
        return caps

    @property
    def client_class(self) -> Type["MotherDuckClient"]:
        from dlt.destinations.impl.motherduck.motherduck import MotherDuckClient

        return MotherDuckClient

    def __init__(
        self,
        credentials: Union[MotherDuckCredentials, str, Dict[str, Any], DuckDBPyConnection] = None,
        create_indexes: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the MotherDuck destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[MotherDuckCredentials, str, Dict[str, Any], DuckDBPyConnection], optional): Credentials to connect to the MotherDuck database. Can be an instance of `MotherDuckCredentials` or
                a connection string in the format `md:///<database_name>?token=<service token>`.
                Instance of `DuckDbCredentials` allows to pass access token, extensions, configs and pragmas to be set up for connection.
            create_indexes (bool, optional): Should unique indexes be created
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            create_indexes=create_indexes,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


motherduck.register()



import io
import contextlib
from typing import Set

from .vault import VaultDocProvider


class AirflowSecretsTomlProvider(VaultDocProvider):
    def __init__(
        self,
        only_secrets: bool = False,
        only_toml_fragments: bool = False,
        list_secrets: bool = False,
    ) -> None:
        super().__init__(only_secrets, only_toml_fragments, list_secrets)

    @property
    def name(self) -> str:
        return "Airflow Secrets TOML Provider"

    def _look_vault(self, full_key: str, hint: type) -> str:
        """Get Airflow Variable with given `full_key`, return None if not found"""
        from airflow.models import Variable

        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            return Variable.get(full_key, default_var=None)  # type: ignore

    def _list_vault(self) -> Set[str]:
        """Lists all available variables in Airflow

        Returns:
            Set[str]: A set of available variable names in Airflow
        """
        from airflow.models import Variable
        from dlt.common import logger

        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            # Use Airflow's session to query all variables
            session = Variable.get_session()
            available_keys = {var.key for var in session.query(Variable.key).all()}
            logger.info(f"Listed {len(available_keys)} variables from Airflow")
            return available_keys

    @property
    def supports_secrets(self) -> bool:
        return True



from abc import abstractmethod
import re
import duckdb
from fsspec import AbstractFileSystem
from packaging.version import Version
from pathlib import Path
import sqlglot
import sqlglot.expressions as exp
from urllib.parse import urlparse
import math
from contextlib import contextmanager
from typing import (
    Any,
    AnyStr,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Generator,
    Type,
    cast,
)

from dlt.common import logger
from dlt.common.configuration.specs.hf_credentials import HfCredentials
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.configuration.specs import (
    AwsCredentials,
    AzureServicePrincipalCredentialsWithoutDefaults,
    AzureCredentialsWithoutDefaults,
)
from dlt.common.destination.client import JobClientBase
from dlt.common.destination.dataset import DBApiCursor

from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.storages.configuration import FileSystemCredentials
from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.destinations.impl.duckdb.exceptions import IcebergViewException
from dlt.destinations.typing import DBApi, DBTransaction, DataFrame, ArrowTable
from dlt.destinations.sql_client import (
    SqlClientBase,
    DBApiCursorImpl,
    raise_database_error,
    raise_open_connection_error,
)

from dlt.destinations.impl.duckdb.configuration import (
    DuckDbBaseCredentials,
    DuckDbClientConfiguration,
    DuckDbCredentials,
)


class DuckDBDBApiCursorImpl(DBApiCursorImpl):
    """Use native duckdb data frame support if available"""

    native_cursor: duckdb.DuckDBPyConnection  # type: ignore
    vector_size: ClassVar[int] = 2048  # vector size is 2048

    def _get_page_count(self, chunk_size: int) -> int:
        """get the page count for vector size"""
        if chunk_size < self.vector_size:
            return 1
        return math.floor(chunk_size / self.vector_size)

    def iter_df(self, chunk_size: int) -> Generator[DataFrame, None, None]:
        # full frame
        if not chunk_size:
            yield self.native_cursor.fetch_df()
            return
        # iterate
        while True:
            df = self.native_cursor.fetch_df_chunk(self._get_page_count(chunk_size))
            if df.shape[0] == 0:
                break
            yield df

    def iter_arrow(self, chunk_size: int) -> Generator[ArrowTable, None, None]:
        if not chunk_size:
            yield self.native_cursor.fetch_arrow_table()
            return
        # iterate
        for item in self.native_cursor.fetch_record_batch(chunk_size):
            yield ArrowTable.from_batches([item])

    def close(self, *args: Any, **kwargs: Any) -> None:
        # duckdb cursor is just original connection so we cannot close it
        pass


class DuckDbSqlClient(SqlClientBase[duckdb.DuckDBPyConnection], DBTransaction):
    dbapi: ClassVar[DBApi] = duckdb
    cursor_impl: ClassVar[Type[DuckDBDBApiCursorImpl]] = DuckDBDBApiCursorImpl

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: DuckDbBaseCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(None, dataset_name, staging_dataset_name, capabilities)
        self._conn: duckdb.DuckDBPyConnection = None
        self.credentials = credentials
        # set additional connection options so derived class can change it
        # TODO: move that to methods that can be overridden, include local_config
        self._pragmas = ["enable_checkpoint_on_shutdown"]
        self._global_config: Dict[str, Any] = {
            "TimeZone": "UTC",
            "checkpoint_threshold": "1gb",
        }

    @raise_open_connection_error
    def open_connection(self) -> duckdb.DuckDBPyConnection:
        self._conn = self.credentials.conn_pool.borrow_conn(
            pragmas=self._pragmas,
            global_config=self._global_config,
            local_config={
                "search_path": self.fully_qualified_dataset_name(),
            },
        )
        return self._conn

    def close_connection(self) -> None:
        if self._conn:
            self.credentials.conn_pool.return_conn(self._conn)
            self._conn = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            self._conn.begin()
            yield self
        except Exception:
            # in some cases duckdb rollback the transaction automatically
            try:
                self.rollback_transaction()
            except DatabaseTransientException:
                pass
            raise
        try:
            self.commit_transaction()
        except Exception:
            try:
                self.rollback_transaction()
            except DatabaseTransientException:
                pass
            raise

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()

    @raise_database_error
    def rollback_transaction(self) -> None:
        self._conn.rollback()

    @property
    def native_connection(self) -> duckdb.DuckDBPyConnection:
        return self._conn

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                return curr.fetchall()

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        assert isinstance(query, str)
        db_args = args if args else kwargs if kwargs else None
        if db_args:
            # TODO: must provide much better refactoring of params
            query = query.replace("%s", "?")

        # we can't duplicate connection because local settings (ie. search path) are lost
        cur = self._conn
        try:
            yield self.cursor_impl(cur.execute(query, db_args))  # type: ignore
        except duckdb.Error:
            try:
                cur.rollback()
            except Exception:
                pass
            raise
        finally:
            # we should somehow cleanup result set if present but no suitable method on connection
            pass

    def warn_if_catalog_equals_dataset_name(self) -> None:
        """
        Checks if the DuckDB connection's current catalog equals the dataset name (schema).
        """
        try:
            # First try to get the catalog via a function that (if available) returns the current database.
            result = self._conn.execute("SELECT current_database()").fetchone()
            if result and len(result) > 0:
                catalog = result[0]
            else:
                # fallback: use PRAGMA database_list to fetch the first (default) database name.
                result = self._conn.execute("PRAGMA database_list").fetchone()
                catalog = result[0] if result else None
        except Exception:
            return

        if catalog is None:
            return

        if catalog == self.dataset_name:
            logger.warning(
                "The current catalog (typically database file name) '%s' is identical to the"
                " dataset name '%s'. This may lead to confusion in the DuckDB binder. Consider"
                " using distinct names. Most typically you use the same name for your pipeline and"
                " dataset or the same name for your destination and the dataset.",
                catalog,
                self.dataset_name,
            )

    def create_secret_name(self, scope: str) -> str:
        regex = re.compile("[^a-zA-Z]")
        escaped_bucket_name = regex.sub("", scope.lower())
        return f"{self.dataset_name}_{escaped_bucket_name}"

    @raise_database_error
    def list_secrets(self) -> Sequence[str]:
        """List secrets that belong to this dataset"""
        secrets = self._conn.sql(
            f"SELECT name FROM duckdb_secrets() WHERE name LIKE '{self.dataset_name}%'"
        ).fetchall()
        return [s[0] for s in secrets]

    @raise_database_error
    def drop_secret(self, secret_name: str) -> None:
        if not secret_name.startswith(self.dataset_name):
            raise ValueError(
                f"Secret name must start with dataset name `{self.dataset_name}`, got"
                f" `{secret_name}`."
            )

        self._conn.sql(f"DROP SECRET {secret_name}")

    @raise_database_error
    def create_secret(
        self,
        scope: str,
        credentials: FileSystemCredentials,
        secret_name: str = None,
        persist_secrets: bool = False,
    ) -> bool:
        #  home dir is a bad choice, it should be more explicit
        if not secret_name:
            secret_name = self.create_secret_name(scope)

        if not secret_name.startswith(self.dataset_name):
            raise ValueError(
                f"Secret name must start with dataset name {self.dataset_name}, got {secret_name}"
            )

        secrets_path = Path(
            self._conn.sql(
                "SELECT current_setting('secret_directory') AS secret_directory"
            ).fetchone()[0]
        )

        is_default_secrets_directory = (
            len(secrets_path.parts) >= 2
            and secrets_path.parts[-1] == "stored_secrets"
            and secrets_path.parts[-2] == ".duckdb"
        )

        if is_default_secrets_directory and persist_secrets:
            logger.warn(
                "You are persisting duckdb secrets but are storing them in the default folder"
                f" {secrets_path}. These secrets are saved there unencrypted, we"
                " recommend using a custom secret directory."
            )

        persistent_stmt = ""
        if persist_secrets:
            persistent_stmt = " PERSISTENT "

        if "@" in scope:
            scope = scope.split("@")[0]

        protocol = urlparse(scope).scheme
        sql: List[str] = []

        # add secrets required for creating views
        if protocol == "s3":
            aws_creds = cast(AwsCredentials, credentials)
            session_token = (
                "" if aws_creds.aws_session_token is None else aws_creds.aws_session_token
            )

            use_ssl = "true"
            endpoint = aws_creds.endpoint_url or "s3.amazonaws.com"
            if aws_creds.endpoint_url and "http://" in aws_creds.endpoint_url:
                use_ssl = "false"
                endpoint = aws_creds.endpoint_url.replace("http://", "")
            elif aws_creds.endpoint_url and "https://" in aws_creds.endpoint_url:
                endpoint = aws_creds.endpoint_url.replace("https://", "")

            s3_url_style = aws_creds.s3_url_style or "path"
            sql.append(f"""
                CREATE OR REPLACE {persistent_stmt} SECRET {secret_name} (
                    TYPE S3,
                    KEY_ID '{aws_creds.aws_access_key_id}',
                    SECRET '{aws_creds.aws_secret_access_key}',
                    SESSION_TOKEN '{session_token}',
                    REGION '{aws_creds.region_name}',
                    ENDPOINT '{endpoint}',
                    SCOPE '{scope}',
                    URL_STYLE '{s3_url_style}',
                    USE_SSL {use_ssl}
                )""")

        # azure with storage account creds
        elif protocol in ["az", "abfss"]:
            # the line below solves problems with certificate path lookup on linux
            # see duckdb docs
            sql.append("SET azure_transport_option_type = 'curl'")

            if isinstance(credentials, AzureCredentialsWithoutDefaults):
                sql.append(f"""
                CREATE OR REPLACE {persistent_stmt} SECRET {secret_name} (
                    TYPE AZURE,
                    CONNECTION_STRING 'AccountName={credentials.azure_storage_account_name};AccountKey={credentials.azure_storage_account_key}',
                    SCOPE '{scope}'
                )""")

            # azure with service principal creds
            elif isinstance(credentials, AzureServicePrincipalCredentialsWithoutDefaults):
                sql.append(f"""
                CREATE OR REPLACE {persistent_stmt} SECRET {secret_name} (
                    TYPE AZURE,
                    PROVIDER SERVICE_PRINCIPAL,
                    TENANT_ID '{credentials.azure_tenant_id}',
                    CLIENT_ID '{credentials.azure_client_id}',
                    CLIENT_SECRET '{credentials.azure_client_secret}',
                    ACCOUNT_NAME '{credentials.azure_storage_account_name}',
                    SCOPE '{scope}'
                )""")

        elif protocol == "hf":
            credentials = cast(HfCredentials, credentials)
            sql.append(f"""
            CREATE OR REPLACE {persistent_stmt} SECRET {secret_name} (
                TYPE HUGGINGFACE,
                TOKEN '{credentials.hf_token}'
            )""")

        elif persist_secrets:
            raise ValueError(
                "Cannot create persistent secret for filesystem protocol"
                f" `{protocol}`. If you are trying to use persistent secrets"
                " with gs/gcs, please use the s3 compatibility layer."
            )
        else:
            # could not create secret
            return False
        self._conn.sql(";\n".join(sql))
        return True

    def use_dataset(self) -> None:
        """Makes duckdb schema corresponding to dataset_name the default"""
        fq_name = self.fully_qualified_dataset_name()
        self._conn.sql(f"SET search_path = '{fq_name}'")

    def _register_filesystem(self, fs: AbstractFileSystem, scheme: str) -> None:
        protocols = [fs.protocol] if isinstance(fs.protocol, str) else fs.protocol
        for protocol in protocols:
            if self._conn.filesystem_is_registered(protocol):
                self._conn.unregister_filesystem(protocol)
        self._conn.register_filesystem(fs)
        if scheme not in protocols:
            fs.protocol = scheme
            if self._conn.filesystem_is_registered(fs.protocol):
                self._conn.unregister_filesystem(fs.protocol)
            self._conn.register_filesystem(fs)

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, duckdb.CatalogException):
            if "already exists" in str(ex):
                return DatabaseTerminalException(ex)
            else:
                return DatabaseUndefinedRelation(ex)
        elif isinstance(ex, duckdb.BinderException) and "not found in DuckLakeCatalog" in str(ex):
            return DatabaseUndefinedRelation(ex)
        elif isinstance(ex, duckdb.InvalidInputException):
            if "Catalog Error" in str(ex):
                return DatabaseUndefinedRelation(ex)
            # duckdb raises TypeError on malformed query parameters
            return DatabaseTransientException(duckdb.ProgrammingError(ex))
        elif isinstance(ex, duckdb.IOException):
            message = str(ex)
            if (
                "delta" in message and "No files in log segment" in message
            ) or "Path does not exist" in message:
                # delta scanner with no delta data and metadata exist in the location
                return DatabaseUndefinedRelation(ex)
            if "Could not guess Iceberg table version" in message:
                # same but iceberg
                return DatabaseUndefinedRelation(ex)
            if "No files found" in message:
                # glob patterns not found
                return DatabaseUndefinedRelation(ex)
            return DatabaseTransientException(ex)
        elif isinstance(ex, duckdb.InternalException):
            if "INTERNAL Error: Value::LIST(values)" in str(ex):
                return IcebergViewException(
                    ex,
                    "duckdb Iceberg extension raises this error when empty (no data) Iceberg table"
                    " is queried. https://github.com/duckdb/duckdb-iceberg/issues/65",
                )
            else:
                return DatabaseTransientException(ex)
        elif isinstance(
            ex,
            (
                duckdb.OperationalError,
                duckdb.InternalError,
                duckdb.SyntaxException,
                duckdb.ParserException,
            ),
        ):
            return DatabaseTransientException(ex)
        elif isinstance(
            ex,
            (
                duckdb.DataError,
                duckdb.ProgrammingError,
                duckdb.IntegrityError,
                duckdb.NotImplementedException,
            ),
        ):
            return DatabaseTerminalException(ex)
        elif cls.is_dbapi_exception(ex):
            return DatabaseTransientException(ex)
        else:
            return ex

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, duckdb.Error)


class WithTableScanners(DuckDbSqlClient):
    memory_db: duckdb.DuckDBPyConnection = None
    """Internally created in-mem database in case external is not provided"""

    def __init__(
        self,
        remote_client: JobClientBase,
        dataset_name: str,
        cache_db: DuckDbCredentials = None,
        persist_secrets: bool = False,
    ) -> None:
        """Allows to maps data in tables accessed via `remote_client` as VIEWs in duckdb database.
        Creates in memory "cache" database by default or allows for external database via "cache_db".
        Will attempt to create views lazily by parsing SQL queries, identifying tables and adding views
        before execution.
        """
        # if no credentials are passed from the outside
        # we know to keep an in memory instance here
        if not cache_db:
            if persist_secrets:
                raise Exception("Creating persistent secrets for in memory db is not allowed.")
            self.memory_db = duckdb.connect(":memory:")
            cache_db = DuckDbCredentials(self.memory_db)
        if not cache_db.is_resolved():
            # create connection pool
            cache_db.resolve()

        from dlt.destinations.impl.duckdb.factory import duckdb as duckdb_factory

        super().__init__(
            dataset_name=dataset_name,
            staging_dataset_name=None,
            credentials=cache_db,
            capabilities=duckdb_factory().capabilities(
                DuckDbClientConfiguration(credentials=cache_db), naming=remote_client.schema.naming
            ),
        )
        self.remote_client = remote_client
        self.schema = remote_client.schema
        self.persist_secrets = persist_secrets
        self._global_config.update(
            {
                "enable_http_metadata_cache": True,
            }
        )

        if Version(duckdb.__version__) >= Version("1.2.0"):
            self._global_config.update(
                {
                    # prevents HEAD command by caching parquet metadata
                    "parquet_metadata_cache": True,
                }
            )

    @raise_database_error
    def open_connection(self) -> duckdb.DuckDBPyConnection:
        # lock the whole process to prevent race condition on `first_connection` flag
        # which may become invalid if second connection got opened in another thread
        with self.credentials.conn_pool._conn_lock:
            first_connection = self.credentials.conn_pool.never_borrowed
            super().open_connection()

        try:
            # NOTE: do not self.execute*** methods when opening connection, may end in endless recursion
            if first_connection:
                # set up dataset
                q_dataset_name = self.fully_qualified_dataset_name()
                create_schema_sql = "CREATE SCHEMA IF NOT EXISTS %s" % q_dataset_name
                self._conn.sql(f"{create_schema_sql};USE {self.fully_qualified_dataset_name()}")
        except Exception:
            self.close_connection()
            raise

        return self._conn

    @abstractmethod
    def should_replace_view(self, view_name: str, table_schema: PreparedTableSchema) -> bool:
        """Tells if view `view_name` should be replaced"""
        pass

    @abstractmethod
    def create_view(self, view_name: str, table_schema: PreparedTableSchema) -> None:
        pass

    @abstractmethod
    def can_create_view(self, table_schema: PreparedTableSchema) -> bool:
        """Tells if a view for a table `table_schema` can be created"""
        pass

    def create_views_for_all_tables(self) -> None:
        self.create_views_for_tables({v: v for v in self.schema.tables.keys()})

    def create_views_for_tables(self, tables: Dict[str, str]) -> None:
        """Add the required tables as views to the duckdb in memory instance"""

        # this also gets all views
        existing_tables = [tname[0] for tname in self._conn.execute("SHOW TABLES").fetchall()]
        # map only tables with data
        tables_with_data = self.schema.dlt_table_names() + self.schema.data_table_names(
            seen_data_only=True
        )

        for table_name in tables.keys():
            view_name = tables[table_name]

            if table_name not in tables_with_data:
                # unknown views will not be created
                continue
            # NOTE: if this is staging configuration then `prepare_load_table` will remove some info
            # from table schema, if we ever extend this to handle staging destination, this needs to change
            schema_table = self.remote_client.prepare_load_table(table_name)

            needs_replace = self.should_replace_view(view_name, schema_table)
            # skip if view already exists and does not need to be replaced each time
            if view_name in existing_tables and not needs_replace:
                continue

            if not self.can_create_view(schema_table):
                continue

            self.create_view(view_name, schema_table)

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        # skip parametrized queries, we could also render them but currently user is not able to
        # do parametrized queries via dataset interface
        if args or kwargs:
            # this is hack
            query = query.replace("%s", "?")  # type: ignore
        # find all tables to preload
        expression = sqlglot.parse_one(query, read="duckdb")  # type: ignore
        load_tables: Dict[str, str] = {}
        for table in expression.find_all(exp.Table):
            # sqlglot has tables without tables ie. schemas are tables
            if not table.this:
                continue
            schema = table.db
            # add only tables from the dataset schema
            if schema or schema.lower() != self.dataset_name.lower():
                load_tables[table.name] = table.name

        if load_tables:
            self.create_views_for_tables(load_tables)
        with super().execute_query(query, *args, **kwargs) as cursor:
            yield cursor

    @staticmethod
    def _setup_iceberg(conn: duckdb.DuckDBPyConnection) -> None:
        if Version(duckdb.__version__) <= Version("1.1.2"):
            raise NotImplementedError(
                f"Iceberg scanner for duckdb `{duckdb.__version__}` does not implement recent"
                " snapshot discovery. Please install duckdb >= 1.1.3"
            )
        # needed to make persistent secrets work in new connection
        # https://github.com/duckdb/duckdb_iceberg/issues/83
        conn.execute("FROM duckdb_secrets()")

        # `duckdb_iceberg` extension does not support autoloading
        # https://github.com/duckdb/duckdb_iceberg/issues/71
        if Version(duckdb.__version__) < Version("1.2.0"):
            conn.execute("INSTALL Iceberg FROM core_nightly; LOAD iceberg")

    def __del__(self) -> None:
        if self.memory_db:
            self.memory_db.close()
            self.memory_db = None



from typing import Any, Callable, Dict, Optional, Sequence

from dlt import version
from dlt.common.exceptions import MissingDependencyException

try:
    from pyiceberg.transforms import (
        Transform,
        IdentityTransform,
        YearTransform,
        MonthTransform,
        DayTransform,
        HourTransform,
        BucketTransform,
        TruncateTransform,
        S,
    )
    from pyiceberg.partitioning import (
        PartitionSpec as IcebergPartitionSpec,
        PartitionField,
        PARTITION_FIELD_ID_START,
    )
    from pyiceberg.schema import Schema as IcebergSchema
    from pyiceberg.io.pyarrow import pyarrow_to_schema
    from pyiceberg.table.name_mapping import NameMapping, MappedField
except ImportError:
    raise MissingDependencyException(
        "dlt iceberg partition spec",
        [f"{version.DLT_PKG_NAME}[pyiceberg]"],
        "Install `pyiceberg` for dlt iceberg partition spec utilities to work",
    )

from dlt.common.libs.pyarrow import pyarrow as pa
from dlt.destinations.impl.filesystem.iceberg_adapter import PartitionSpec


_TRANSFORM_LOOKUP: Dict[str, Callable[[Optional[int]], Transform[S, Any]]] = {
    "identity": lambda _: IdentityTransform(),
    "year": lambda _: YearTransform(),
    "month": lambda _: MonthTransform(),
    "day": lambda _: DayTransform(),
    "hour": lambda _: HourTransform(),
    "bucket": lambda n: BucketTransform(n),
    "truncate": lambda w: TruncateTransform(w),
}


def get_partition_transform(spec: PartitionSpec) -> Transform[S, Any]:
    """Get the PyIceberg Transform object for a partition spec.

    Args:
        spec: The PartitionSpec to get the transform for

    Returns:
        A PyIceberg Transform object

    Raises:
        ValueError: If the transform is not recognized
    """
    try:
        factory = _TRANSFORM_LOOKUP[spec.transform]
    except KeyError as exc:
        raise ValueError(f"Unknown partition transformation type: {spec.transform}") from exc
    return factory(spec.param_value)


def build_iceberg_partition_spec(
    arrow_schema: pa.Schema,
    spec_list: Sequence[PartitionSpec],
) -> tuple[IcebergPartitionSpec, IcebergSchema]:
    """
    Turn a dlt PartitionSpec list into a PyIceberg PartitionSpec.
    Returns the PartitionSpec and the IcebergSchema derived from the Arrow schema.
    """
    name_mapping = NameMapping(
        [
            MappedField(field_id=i + 1, names=[name])  # type: ignore[call-arg]
            for i, name in enumerate(arrow_schema.names)
        ]
    )
    iceberg_schema: IcebergSchema = pyarrow_to_schema(arrow_schema, name_mapping)

    fields: list[PartitionField] = []
    for pos, spec in enumerate(spec_list):
        iceberg_field = iceberg_schema.find_field(spec.source_column)

        fields.append(
            PartitionField(
                field_id=PARTITION_FIELD_ID_START + pos,
                source_id=iceberg_field.field_id,
                transform=get_partition_transform(spec),
                name=_default_field_name(spec),
            )
        )

    return IcebergPartitionSpec(*fields), iceberg_schema


def _default_field_name(spec: PartitionSpec) -> str:
    """
    Replicate Iceberg's automatic partition-field naming by delegating to the private
    _PartitionNameGenerator. Falls back to the user-supplied `partition_field` if present.
    """
    from pyiceberg.partitioning import _PartitionNameGenerator

    name_generator = _PartitionNameGenerator()

    if spec.partition_field:  # user-supplied `partition_field`
        return spec.partition_field

    # name generator requires field_id and source_id, but does not use them
    dummy_field_id = 0
    dummy_source_id = 0

    if spec.transform == "bucket":
        # bucket / truncate need the numeric parameter
        return name_generator.bucket(
            dummy_field_id, spec.source_column, dummy_source_id, spec.param_value
        )
    if spec.transform == "truncate":
        return name_generator.truncate(
            dummy_field_id, spec.source_column, dummy_source_id, spec.param_value
        )

    # identity, year, month, day, hour – all have the same signature
    method = getattr(name_generator, spec.transform)

    return method(dummy_field_id, spec.source_column, dummy_source_id)  # type: ignore[no-any-return]



from dlt.common.destination.exceptions import DestinationException, DestinationTerminalException


class WeaviateBatchError(DestinationException):
    """Error raised when batch operations fail in Weaviate"""

    pass


# Keep old name for backwards compatibility
WeaviateGrpcError = WeaviateBatchError


class PropertyNameConflict(DestinationTerminalException):
    def __init__(self, error: str) -> None:
        super().__init__(
            "Your data contains items with identical property names when compared case insensitive."
            " Weaviate cannot handle such data. Please clean up your data before loading or change"
            " to case insensitive naming convention. See"
            " https://dlthub.com/docs/dlt-ecosystem/destinations/weaviate#names-normalization for"
            f" details. [{error}]"
        )



import abc
from typing import Any, Sequence, Tuple, Type, Optional

from dlt.common.utils import simple_repr, without_none

EXPLICIT_VALUES_PROVIDER_NAME = "ExplicitValues"
"""A name of virtual provider that represent explicit values found during config resolution"""


class ConfigProvider(abc.ABC):
    @abc.abstractmethod
    def get_value(
        self, key: str, hint: Type[Any], pipeline_name: str, *sections: str
    ) -> Tuple[Optional[Any], str]:
        """Looks for a value under `key` in section(s) `sections` and tries to coerce the
        value to type `hint`. A pipeline context (top level section) will be added if
        `pipeline_name` was specified.
        """

    def set_value(self, key: str, value: Any, pipeline_name: str, *sections: str) -> None:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def supports_secrets(self) -> bool:
        """If true, provider is allowed to store secret. Configuration resolution fails if
        a secret value is discovered in a config provider that does not support secrets.
        """

    @property
    @abc.abstractmethod
    def supports_sections(self) -> bool:
        """If true, config resolution will query this provider for all allowed section combinations
        otherwise values are queried only by field name.
        """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Human readable name of config provider"""

    @property
    def is_empty(self) -> bool:
        """Tells if config provider holds any values"""
        return False

    @property
    def is_writable(self) -> bool:
        """Tells if `set_value` may be used"""
        return False

    @property
    def locations(self) -> Sequence[str]:
        """Returns a all possible locations where secrets may be stored, human readable"""
        return []

    @property
    def present_locations(self) -> Sequence[str]:
        """Returns a list of locations that were present and contained secrets, human readable"""
        return self.locations

    def __repr__(self) -> str:
        kwargs = {
            "is_empty": self.is_empty,
            "supports_secrets": self.supports_secrets,
            "supports_sections": self.supports_sections,
            "is_writable": self.is_writable,
            "locations": self.locations if self.locations else None,
        }
        return simple_repr(self.__class__.__name__, **without_none(kwargs))


def get_key_name(key: str, separator: str, /, *sections: str) -> str:
    if sections:
        sections = filter(lambda x: bool(x), sections)  # type: ignore
        env_key = separator.join((*sections, key))
    else:
        env_key = key
    return env_key



import datetime  # noqa: I251
import re

from clickhouse_driver import dbapi as clickhouse_dbapi  # type: ignore[import-untyped]
import clickhouse_driver
import clickhouse_driver.errors  # type: ignore[import-untyped]
from clickhouse_driver.dbapi import OperationalError  # type: ignore[import-untyped]
from clickhouse_driver.dbapi.extras import DictCursor  # type: ignore[import-untyped]
import clickhouse_connect
from clickhouse_connect.driver.tools import insert_file as clk_insert_file
from clickhouse_connect.driver.summary import QuerySummary

from contextlib import contextmanager
from typing import (
    Iterator,
    AnyStr,
    Any,
    List,
    Optional,
    Sequence,
    ClassVar,
    Literal,
    Tuple,
    cast,
)

from pendulum import DateTime  # noqa: I251

from dlt.common import logger
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.time import ensure_pendulum_datetime_non_utc, normalize_timezone
from dlt.common.typing import DictStrAny
from dlt.common.utils import removeprefix

from dlt.destinations.exceptions import (
    DatabaseUndefinedRelation,
    DatabaseTransientException,
    DatabaseTerminalException,
)
from dlt.destinations.impl.clickhouse.configuration import (
    ClickHouseCredentials,
    ClickHouseClientConfiguration,
)
from dlt.destinations.impl.clickhouse.typing import (
    TTableEngineType,
    TABLE_ENGINE_TYPE_TO_CLICKHOUSE_ATTR,
)
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    raise_database_error,
    raise_open_connection_error,
)
from dlt.destinations.typing import DBTransaction, DBApi
from dlt.destinations.utils import _convert_to_old_pyformat


TDeployment = Literal["ClickHouseOSS", "ClickHouseCloud"]
TRANSACTIONS_UNSUPPORTED_WARNING_MESSAGE = (
    "ClickHouse does not support transactions! Each statement is auto-committed separately."
)


class ClickHouseDBApiCursorImpl(DBApiCursorImpl):
    native_cursor: DictCursor

    # TODO: implement arrow reading


class ClickHouseSqlClient(
    SqlClientBase[clickhouse_driver.dbapi.connection.Connection], DBTransaction
):
    dbapi: ClassVar[DBApi] = clickhouse_dbapi

    def __init__(
        self,
        dataset_name: Optional[str],
        staging_dataset_name: str,
        known_table_names: List[str],
        credentials: ClickHouseCredentials,
        capabilities: DestinationCapabilitiesContext,
        config: ClickHouseClientConfiguration,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self._conn: clickhouse_driver.dbapi.connection = None
        self.known_table_names = known_table_names
        self.credentials = credentials
        self.database_name = credentials.database
        self.config = config

    def has_dataset(self) -> bool:
        # we do not need to normalize dataset_sentinel_table_name.
        sentinel_table = self.config.dataset_sentinel_table_name
        all_ds_tables = self._list_tables()
        if self.dataset_name:
            prefix = self.dataset_name + self.config.dataset_table_separator
            return sentinel_table in [removeprefix(t, prefix) for t in all_ds_tables]
        else:
            # if no dataset specified we look for sentinel table
            return sentinel_table in all_ds_tables

    def open_connection(self) -> clickhouse_driver.dbapi.connection.Connection:
        if self.credentials.query is None:
            self.credentials.query = {}
        self.credentials.query["select_sequential_consistency"] = (
            self.config.select_sequential_consistency
        )
        self._conn = clickhouse_driver.connect(dsn=self.credentials.to_native_representation())
        return self._conn

    @raise_open_connection_error
    def close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        yield self

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()

    @raise_database_error
    def rollback_transaction(self) -> None:
        self._conn.rollback()

    @property
    def native_connection(self) -> clickhouse_driver.dbapi.connection.Connection:
        return self._conn

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            return None if curr.description is None else curr.fetchall()

    def create_dataset(self) -> None:
        # We create a sentinel table which defines whether we consider the dataset created.
        sentinel_table_name = self.make_qualified_table_name(
            self.config.dataset_sentinel_table_name
        )
        sentinel_table_type = cast(TTableEngineType, self.config.table_engine_type)
        self.execute_sql(f"""
            CREATE TABLE {sentinel_table_name}
            (_dlt_id String NOT NULL)
            ENGINE={TABLE_ENGINE_TYPE_TO_CLICKHOUSE_ATTR.get(sentinel_table_type)}
            PRIMARY KEY _dlt_id
            COMMENT 'internal dlt sentinel table'""")

    def drop_dataset(self) -> None:
        # always try to drop the sentinel table.
        sentinel_table_name = self.make_qualified_table_name(
            self.config.dataset_sentinel_table_name
        )

        all_ds_tables = self._list_tables()

        if self.dataset_name:
            # Since ClickHouse doesn't have schemas, we need to drop all tables in our virtual schema,
            # or collection of tables, that has the `dataset_name` as a prefix.
            to_drop_results = all_ds_tables
        else:
            # drop only tables known in logical (dlt) schema
            to_drop_results = [
                table_name for table_name in self.known_table_names if table_name in all_ds_tables
            ]

        catalog_name = self.catalog_name()
        # drop a sentinel table only when dataset name was empty (was not included in the schema)
        if not self.dataset_name:
            self.execute_sql(f"DROP TABLE {sentinel_table_name} SYNC")
            logger.warning(
                "Dataset without name (tables without prefix) got dropped. Only tables known in the"
                " current dlt schema and sentinel tables were removed."
            )
        else:
            sentinel_table_name = self.make_qualified_table_name_path(
                self.config.dataset_sentinel_table_name, quote=False
            )[-1]
            if sentinel_table_name not in all_ds_tables:
                # no sentinel table, dataset does not exist
                self.execute_sql(f"SELECT 1 FROM {sentinel_table_name}")
                raise AssertionError(f"{sentinel_table_name} must not exist")
        for table in to_drop_results:
            # The "DROP TABLE" clause is discarded if we allow clickhouse_driver to handle parameter substitution.
            # This is because the driver incorrectly substitutes the entire query string, causing the "DROP TABLE" keyword to be omitted.
            # To resolve this, we are forced to provide the full query string here.
            self.execute_sql(
                f"DROP TABLE {catalog_name}.{self.capabilities.escape_identifier(table)} SYNC"
            )

    def drop_tables(self, *tables: str) -> None:
        """Drops a set of tables if they exist"""
        if not tables:
            return
        statements = [
            f"DROP TABLE IF EXISTS {self.make_qualified_table_name(table)} SYNC" for table in tables
        ]
        self.execute_many(statements)

    def insert_file(
        self, file_path: str, table_name: str, file_format: str, compression: str
    ) -> QuerySummary:
        with clickhouse_connect.create_client(
            host=self.credentials.host,
            port=self.credentials.http_port,
            database=self.credentials.database,
            user_name=self.credentials.username,
            password=self.credentials.password,
            secure=bool(self.credentials.secure),
        ) as clickhouse_connect_client:
            return clk_insert_file(
                clickhouse_connect_client,
                self.make_qualified_table_name(table_name),
                file_path,
                fmt=file_format,
                settings={
                    "allow_experimental_lightweight_delete": 1,
                    "enable_http_compression": 1,
                    "date_time_input_format": "best_effort",
                    "select_sequential_consistency": self.config.select_sequential_consistency,
                },
                compression=compression,
            )

    def _list_tables(self) -> List[str]:
        catalog_name, table_name = self.make_qualified_table_name_path("%", quote=False)
        rows = self.execute_sql(
            """
            SELECT name
            FROM system.tables
            WHERE database = %s
            AND name LIKE %s
            """,
            catalog_name,
            table_name,
        )
        return [row[0] for row in rows]

    @staticmethod
    def _sanitise_dbargs(db_args: DictStrAny) -> DictStrAny:
        """For ClickHouse OSS, the DBapi driver doesn't parse datetime types.

        Converts datetime values to naive UTC strings preserving microsecond precision.
        """
        for key, value in db_args.items():
            if isinstance(value, (DateTime, datetime.datetime)):
                value = ensure_pendulum_datetime_non_utc(value)
                db_args[key] = str(normalize_timezone(value, timezone=False))
        return db_args

    @contextmanager
    @raise_database_error
    def execute_query(
        self, query: AnyStr, *args: Any, **kwargs: Any
    ) -> Iterator[ClickHouseDBApiCursorImpl]:
        assert isinstance(query, str), "Query must be a string."

        db_args: DictStrAny = kwargs.copy()

        if args:
            query, db_args = _convert_to_old_pyformat(query, args, OperationalError)
            db_args.update(kwargs)

        db_args = self._sanitise_dbargs(db_args)

        with self._conn.cursor() as cursor:
            for query_line in query.split(";"):
                if query_line := self.escape_pct(query_line.strip()):
                    try:
                        cursor.execute(query_line, db_args)
                    except KeyError as e:
                        raise DatabaseTransientException(OperationalError()) from e

            yield ClickHouseDBApiCursorImpl(cursor)

    def catalog_name(self, quote: bool = True, casefold: bool = True) -> Optional[str]:
        if casefold:
            database_name = self.capabilities.casefold_identifier(self.database_name)
        else:
            database_name = self.database_name
        if quote:
            database_name = self.capabilities.escape_identifier(database_name)
        return database_name

    def make_qualified_table_name_path(
        self, table_name: Optional[str], quote: bool = True, casefold: bool = True
    ) -> List[str]:
        # get catalog and dataset
        path = super().make_qualified_table_name_path(None, quote=quote, casefold=casefold)
        if table_name:
            # table name combines dataset name and table name
            if self.dataset_name:
                table_name = f"{self.dataset_name}{self.config.dataset_table_separator}{table_name}"
            else:
                # without dataset just use the table name
                pass
            if casefold:
                table_name = self.capabilities.casefold_identifier(table_name)
            if quote:
                table_name = self.capabilities.escape_identifier(table_name)
            # we have only two path components
            path[1] = table_name
        else:
            # we have only one path component, dataset name is included in the table name
            path.pop()
        return path

    def _get_information_schema_components(self, *tables: str) -> Tuple[str, str, List[str]]:
        components = super()._get_information_schema_components(*tables)
        # clickhouse has a catalogue and no schema but uses catalogue as a schema to query the information schema 🤷
        # so we must disable catalogue search. also note that table name is prefixed with logical "dataset_name"
        return (None, components[0], components[2])

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, clickhouse_driver.dbapi.errors.OperationalError):
            # https://github.com/ClickHouse/ClickHouse/blob/master/src/Common/ErrorCodes.cpp
            TERMINAL_CODES = {44, 47, 57, 82}
            UNDEFINED_RELATION_CODES = {60, 81}

            match = re.match(r"Code: (\d+)\.", str(ex))
            if match:
                code = int(match.group(1))
                if code in TERMINAL_CODES:
                    return DatabaseTerminalException(ex)
                elif code in UNDEFINED_RELATION_CODES:
                    return DatabaseUndefinedRelation(ex)

            return DatabaseTransientException(ex)
        elif isinstance(ex, clickhouse_driver.dbapi.errors.InternalError):
            return DatabaseTransientException(ex)
        elif isinstance(
            ex,
            (
                clickhouse_driver.dbapi.errors.DataError,
                clickhouse_driver.dbapi.errors.ProgrammingError,
                clickhouse_driver.dbapi.errors.IntegrityError,
            ),
        ):
            return DatabaseTerminalException(ex)
        elif cls.is_dbapi_exception(ex):
            return DatabaseTransientException(ex)
        else:
            return ex

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, clickhouse_driver.dbapi.Error)

    @staticmethod
    def escape_pct(expr: str) -> str:
        """Returns SQL expression with % characters escaped.

        This doubles % characters used as modulo, wildcard, or literal, but leaves those used in
        placeholders (%s) alone:
        - modulo: 16 % 4  ➜  16 %% 4
        - wildcard: 'test' LIKE '%es%'  ➜  'test' LIKE '%%es%%'
        - literal: '100% sure'  ➜  '100%% sure'
        - placeholder: SELECT %s AS value  ➜  SELECT %s AS value
        """

        return re.sub(r"%(?!\()", "%%", expr)



import binascii
import base64
import dataclasses
import datetime  # noqa: I251
from collections.abc import Mapping as C_Mapping, Sequence as C_Sequence
from functools import lru_cache
from typing import Any, Type, Union, Callable, Dict, Tuple
from enum import Enum

from dlt.common.json import custom_pua_remove, json
from dlt.common.json._simplejson import custom_encode as json_custom_encode
from dlt.common.typing import copy_sig
from dlt.common.wei import Wei
from dlt.common.arithmetics import InvalidOperation, Decimal
from dlt.common.data_types.typing import TDataType
from dlt.common.time import (
    ensure_pendulum_datetime_non_utc,
    ensure_pendulum_date,
    ensure_pendulum_time,
)
from dlt.common.utils import map_nested_values_in_place, str2bool
from dlt.common.pendulum import pendulum
from pendulum import (  # noqa: I251
    DateTime as PendulumDateTime,
    Date as PendulumDate,
    Time as PendulumTime,
)


# fast type map for common Python types to schema types.
PY_TYPE_TO_SC_TYPE: Dict[Type[Any], TDataType] = {
    str: "text",
    int: "bigint",
    float: "double",
    bool: "bool",
    dict: "json",
    list: "json",
    datetime.datetime: "timestamp",
    datetime.date: "date",
    datetime.time: "time",
    PendulumDateTime: "timestamp",
    PendulumDate: "date",
    PendulumTime: "time",
}


def _py_type_to_sc_type(t: Type[Any]) -> TDataType:
    if result := PY_TYPE_TO_SC_TYPE.get(t):
        return result

    # wei is subclass of decimal and must be checked first
    if issubclass(t, (dict, list)):
        return "json"
    if issubclass(t, Wei):
        return "wei"
    if issubclass(t, Decimal):
        return "decimal"
    # datetime is subclass of date and must be checked first
    if issubclass(t, datetime.datetime):
        return "timestamp"
    if issubclass(t, datetime.date):
        return "date"
    if issubclass(t, datetime.time):
        return "time"
    # subclassed basic types (bool cannot be subclassed)
    if issubclass(t, str):
        return "text"
    if issubclass(t, float):
        return "double"
    if issubclass(t, int):
        return "bigint"
    if issubclass(t, bytes):
        return "binary"
    if dataclasses.is_dataclass(t) or issubclass(t, (C_Mapping, C_Sequence)):
        return "json"
    # Enum is coerced to str or int respectively
    if issubclass(t, Enum):
        if issubclass(t, int):
            return "bigint"
        else:
            # str subclass and unspecified enum type translates to text
            return "text"

    raise TypeError(t)


# preserves original signature which lru_cache destroys
py_type_to_sc_type = copy_sig(_py_type_to_sc_type)(lru_cache(maxsize=None)(_py_type_to_sc_type))


def json_to_str(value: Any) -> str:
    return json.dumps(map_nested_values_in_place(custom_pua_remove, value))


def coerce_from_date_types(
    to_type: TDataType, value: datetime.datetime
) -> Union[datetime.datetime, datetime.date, datetime.time, int, float, str]:
    v = ensure_pendulum_datetime_non_utc(value)
    if to_type == "timestamp":
        return v
    if to_type == "text":
        return v.isoformat()
    if to_type == "bigint":
        return v.int_timestamp
    if to_type == "double":
        return v.timestamp()
    if to_type == "date":
        return ensure_pendulum_date(v)
    if to_type == "time":
        return v.time()
    raise TypeError(f"Cannot convert timestamp to `{to_type}`")


def _text_to_binary(value: str) -> bytes:
    if value.startswith("0x"):
        return bytes.fromhex(value[2:])
    try:
        return base64.b64decode(value, validate=True)
    except binascii.Error:
        raise ValueError(value)


def _text_to_bigint(value: str) -> int:
    trim_value = value.strip()
    if trim_value.startswith("0x"):
        return int(trim_value[2:], 16)
    return int(trim_value)


def _text_to_double(value: str) -> float:
    trim_value = value.strip()
    if trim_value.startswith("0x"):
        return float(int(trim_value[2:], 16))
    return float(trim_value)


def _text_to_decimal(value: str) -> Decimal:
    trim_value = value.strip()
    if trim_value.startswith("0x"):
        return Decimal(int(trim_value[2:], 16))
    try:
        return Decimal(trim_value)
    except InvalidOperation:
        raise ValueError(trim_value)


def _text_to_wei(value: str) -> Wei:
    trim_value = value.strip()
    if trim_value.startswith("0x"):
        return Wei(int(trim_value[2:], 16))
    try:
        return Wei(trim_value)
    except InvalidOperation:
        raise ValueError(trim_value)


def _numeric_to_bigint(value: Any) -> int:
    if value % 1 != 0:
        raise ValueError(value)
    return int(value)


def _text_fallback(value: Any) -> str:
    try:
        return str(json_custom_encode(value))
    except TypeError:
        return str(value)


def _text_to_date_fast(value: str) -> PendulumDate:
    """Fast date parsing: native Python + pendulum Date."""
    try:
        d = datetime.date.fromisoformat(value)
        return PendulumDate(d.year, d.month, d.day)
    except Exception:
        return ensure_pendulum_date(value)


_COERCE_DISPATCH: Dict[Tuple[TDataType, TDataType], Callable[[Any], Any]] = {
    # to text
    ("text", "json"): json_to_str,
    ("text", "bigint"): _text_fallback,
    ("text", "double"): _text_fallback,
    ("text", "wei"): _text_fallback,
    ("text", "decimal"): _text_fallback,
    ("text", "bool"): _text_fallback,
    ("text", "binary"): _text_fallback,
    ("text", "timestamp"): _text_fallback,
    ("text", "date"): _text_fallback,
    ("text", "time"): _text_fallback,
    # to binary
    ("binary", "text"): _text_to_binary,
    ("binary", "bigint"): lambda v: v.to_bytes((v.bit_length() + 7) // 8, "little"),
    # to bigint
    ("bigint", "text"): _text_to_bigint,
    ("bigint", "wei"): _numeric_to_bigint,
    ("bigint", "decimal"): _numeric_to_bigint,
    ("bigint", "double"): _numeric_to_bigint,
    # to double
    ("double", "text"): _text_to_double,
    ("double", "bigint"): float,
    ("double", "wei"): float,
    ("double", "decimal"): float,
    # to decimal
    ("decimal", "text"): _text_to_decimal,
    ("decimal", "bigint"): Decimal,
    ("decimal", "double"): Decimal,
    ("decimal", "wei"): Decimal,
    # to wei
    ("wei", "text"): _text_to_wei,
    ("wei", "bigint"): Wei,
    ("wei", "double"): Wei,
    ("wei", "decimal"): Wei,
    # to bool
    ("bool", "text"): str2bool,
    ("bool", "bigint"): bool,
    ("bool", "double"): bool,
    ("bool", "decimal"): bool,
    ("bool", "wei"): bool,
    # to json
    ("json", "text"): json.loads,
    # to timestamp
    ("timestamp", "text"): ensure_pendulum_datetime_non_utc,
    ("timestamp", "bigint"): pendulum.from_timestamp,
    ("timestamp", "double"): pendulum.from_timestamp,
    ("timestamp", "date"): ensure_pendulum_datetime_non_utc,
    # to date (fast path)
    ("date", "text"): _text_to_date_fast,
    ("date", "timestamp"): ensure_pendulum_date,
    ("date", "bigint"): ensure_pendulum_date,
    ("date", "double"): ensure_pendulum_date,
    # to time
    ("time", "text"): ensure_pendulum_time,
}


def coerce_value(to_type: TDataType, from_type: TDataType, value: Any) -> Any:
    if to_type == from_type:
        if to_type == "json":
            return map_nested_values_in_place(custom_pua_remove, value)
        if hasattr(value, "value"):
            if to_type == "text":
                return str(value.value)
            elif to_type == "bigint":
                return int(value.value)
        return value

    # dispatch table lookup
    coercer = _COERCE_DISPATCH.get((to_type, from_type))
    if coercer is not None:
        try:
            return coercer(value)
        except OverflowError as e:
            raise ValueError(value) from e

    raise ValueError(value)



from copy import deepcopy
from textwrap import dedent
from typing import Any, Literal, Optional, List, Sequence, cast
from urllib.parse import ParseResult, urlparse

import clickhouse_connect

from dlt.common.configuration.specs import (
    CredentialsConfiguration,
    AzureCredentialsWithoutDefaults,
    AwsCredentialsWithoutDefaults,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    PreparedTableSchema,
    SupportsStagingDestination,
    HasFollowupJobs,
    RunnableLoadJob,
    FollowupJobRequest,
    LoadJob,
)
from dlt.common.schema import Schema, TColumnSchema
from dlt.common.schema.exceptions import SchemaCorruptedException
from dlt.common.schema.typing import TColumnType
from dlt.common.schema.utils import get_columns_names_with_prop, is_nullable_column
from dlt.common.storages import FileStorage
from dlt.common.storages.configuration import FilesystemConfiguration, ensure_canonical_az_url
from dlt.common.storages.fsspec_filesystem import AZURE_BLOB_STORAGE_PROTOCOLS
from dlt.destinations.exceptions import LoadJobTerminalException
from dlt.destinations.impl.clickhouse.configuration import (
    ClickHouseClientConfiguration,
)
from dlt.destinations.impl.clickhouse.sql_client import ClickHouseSqlClient
from dlt.destinations.impl.clickhouse.typing import (
    CODEC_HINT,
    HINT_TO_CLICKHOUSE_ATTR,
    PARTITION_HINT,
    SETTINGS_HINT,
    SORT_HINT,
    TABLE_ENGINE_TYPE_TO_CLICKHOUSE_ATTR,
    TMergeTreeSettings,
    TMergeTreeSettingsValue,
)
from dlt.destinations.impl.clickhouse.typing import (
    TTableEngineType,
    TABLE_ENGINE_TYPE_HINT,
    FILE_FORMAT_TO_TABLE_FUNCTION_MAPPING,
    SUPPORTED_FILE_FORMATS,
)
from dlt.destinations.impl.clickhouse.utils import (
    convert_storage_to_http_scheme,
)
from dlt.common.data_writers.escape import escape_clickhouse_literal
from dlt.destinations.job_client_impl import (
    SqlJobClientBase,
    SqlJobClientWithStagingDataset,
)
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.utils import get_deterministic_temp_table_name
from dlt.destinations.path_utils import get_file_format_and_compression


class ClickHouseLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(
        self,
        file_path: str,
        config: ClickHouseClientConfiguration,
        staging_credentials: Optional[CredentialsConfiguration] = None,
    ) -> None:
        super().__init__(file_path)
        self._job_client: "ClickHouseClient" = None
        self._staging_credentials = staging_credentials
        self._config = config

    def _get_table_function(
        self,
        *,
        bucket_scheme: str,
        bucket_url: ParseResult,
        compression: str,
        clickhouse_format: str,
    ) -> str:
        """Creates the ClickHouse table function for loading from cloud storage.

        Args:
            bucket_scheme: The scheme of the bucket URL (e.g., 's3', 'gs', 'azure').
            bucket_url: The parsed URL of the bucket.
            compression: The compression type to use ('auto', 'gz', 'none').
            clickhouse_format: The ClickHouse format to use for loading data.

        Returns:
            The ClickHouse table function string.

        Raises:
            LoadJobTerminalException: If the bucket scheme is not supported or if
                credentials are missing.
        """
        if bucket_scheme in ("s3", "gs", "gcs"):
            if not isinstance(self._staging_credentials, AwsCredentialsWithoutDefaults):
                raise LoadJobTerminalException(
                    self._file_path,
                    dedent(
                        """
                        Google Cloud Storage buckets must be configured using the S3 compatible access pattern.
                        Please provide the necessary S3 credentials (access key ID and secret access key), to access the GCS bucket through the S3 API.
                        Refer to https://dlthub.com/docs/dlt-ecosystem/destinations/filesystem#using-s3-compatible-storage.
                    """,
                    ).strip(),
                )

            bucket_http_url = convert_storage_to_http_scheme(
                bucket_url,
                endpoint=self._staging_credentials.endpoint_url,
                use_https=self._config.staging_use_https,
            )
            access_key_id = self._staging_credentials.aws_access_key_id
            secret_access_key = self._staging_credentials.aws_secret_access_key
            auth = "NOSIGN"
            if self._config.credentials.s3_extra_credentials:
                # use extra credentials for S3 compatible storage
                # https://clickhouse.com/docs/sql-reference/table-functions/s3#using-s3-credentials-clickhouse-cloud
                extra_credential_args = [
                    f"{k} = {escape_clickhouse_literal(v)}"
                    for k, v in self._config.credentials.s3_extra_credentials.items()
                ]
                auth = f"extra_credentials({', '.join(extra_credential_args)})"
            elif access_key_id and secret_access_key:
                auth = f"'{access_key_id}','{secret_access_key}'"

            return f"s3('{bucket_http_url}',{auth},'{clickhouse_format}','auto','{compression}')"

        elif bucket_scheme in AZURE_BLOB_STORAGE_PROTOCOLS:
            if not isinstance(self._staging_credentials, AzureCredentialsWithoutDefaults):
                raise LoadJobTerminalException(
                    self._file_path,
                    "Unsigned Azure Blob Storage access from ClickHouse isn't supported as yet.",
                )

            # Authenticated access.
            account_name = self._staging_credentials.azure_storage_account_name
            account_host = (
                self._staging_credentials.azure_account_host
                or f"{account_name}.blob.core.windows.net"
            )
            storage_account_url = ensure_canonical_az_url("", "https", account_name, account_host)
            account_key = self._staging_credentials.azure_storage_account_key

            # build table func
            return f"azureBlobStorage('{storage_account_url}','{bucket_url.netloc}','{bucket_url.path}','{account_name}','{account_key}','{clickhouse_format}','{compression}')"

        else:
            raise LoadJobTerminalException(
                self._file_path,
                f"ClickHouse loader does not support `{bucket_scheme}` filesystem.",
            )

    def run(self) -> None:
        client = self._job_client.sql_client

        bucket_path = None
        bucket_url = None
        bucket_scheme = None
        file_name = self._file_name

        if ReferenceFollowupJobRequest.is_reference_job(self._file_path):
            bucket_path = ReferenceFollowupJobRequest.resolve_reference(self._file_path)
            file_name = FileStorage.get_file_name_from_file_path(bucket_path)
            bucket_url = urlparse(bucket_path)
            bucket_scheme = bucket_url.scheme

        compression = "auto"

        file_format, is_compressed = get_file_format_and_compression(file_name)
        clickhouse_format: str = FILE_FORMAT_TO_TABLE_FUNCTION_MAPPING[
            cast(SUPPORTED_FILE_FORMATS, file_format)
        ]

        if file_format == "jsonl":
            # Auto does not work for jsonl. So we set it to 'none' if not compressed
            compression = "gz" if is_compressed else "none"

        # Don't use the DBAPI driver for local files.
        if not bucket_path or bucket_scheme == "file":
            file_path = (
                self._file_path
                if not bucket_path
                else FilesystemConfiguration.make_local_path(bucket_path)
            )
            try:
                client.insert_file(file_path, self.load_table_name, clickhouse_format, compression)
            except clickhouse_connect.driver.exceptions.Error as e:
                raise LoadJobTerminalException(
                    self._file_path,
                    f"ClickHouse connection failed due to `{e}`.",
                ) from e
            return

        qualified_table_name = client.make_qualified_table_name(self.load_table_name)

        if not (bucket_scheme and bucket_url):
            raise LoadJobTerminalException(
                self._file_path, "Cannot determine bucket scheme and URL from the file path."
            )

        table_function = self._get_table_function(
            bucket_scheme=bucket_scheme,
            bucket_url=bucket_url,
            compression=compression,
            clickhouse_format=clickhouse_format,
        )

        statement = f"INSERT INTO {qualified_table_name} SELECT * FROM {table_function}"
        with client.begin_transaction():
            client.execute_sql(statement)


class ClickHouseMergeJob(SqlMergeFollowupJob):
    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlClientBase[Any]) -> str:
        # Deterministic name for crash recovery - orphaned temp tables can be identified and cleaned up
        with sql_client.with_staging_dataset():
            return sql_client.make_qualified_table_name(
                cls._shorten_table_name(
                    get_deterministic_temp_table_name(table_name, op), sql_client
                )
            )

    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        # Use CREATE OR REPLACE to avoid slow DROP TABLE ... SYNC on replicated ClickHouse clusters.
        # The Atomic database engine (default since ClickHouse 20.5) handles this via atomic swap
        # using renameat2(), which is nearly instant. The old table is cleaned up asynchronously.
        # See: https://github.com/dlt-hub/dlt/issues/3562
        return (
            f"CREATE OR REPLACE TABLE {temp_table_name} ENGINE ="
            f" MergeTree PRIMARY KEY {unique_column} AS {select_sql}"
        )

    @classmethod
    def gen_key_table_clauses(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        key_clauses: Sequence[str],
        for_delete: bool,
    ) -> List[str]:
        join_conditions = " OR ".join([c.format(d="d", s="s") for c in key_clauses])
        return [
            f"FROM {root_table_name} AS d JOIN {staging_root_table_name} AS s ON {join_conditions}"
        ]

    @classmethod
    def gen_update_table_prefix(cls, table_name: str) -> str:
        return f"ALTER TABLE {table_name} UPDATE"

    @classmethod
    def requires_temp_table_for_delete(cls) -> bool:
        return True


class ClickHouseClient(SqlJobClientWithStagingDataset, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: ClickHouseClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = SqlJobClientWithStagingDataset.create_dataset_names(
            schema, config
        )
        self.sql_client: ClickHouseSqlClient = ClickHouseSqlClient(
            dataset_name,
            staging_dataset_name,
            list(schema.tables.keys()),
            config.credentials,
            capabilities,
            config,
        )
        super().__init__(schema, config, self.sql_client)
        self.config: ClickHouseClientConfiguration = config
        self.active_hints = deepcopy(HINT_TO_CLICKHOUSE_ATTR)
        self.type_mapper = self.capabilities.get_type_mapper()

    def prepare_load_table(self, table_name: str) -> Optional[PreparedTableSchema]:
        table = super().prepare_load_table(table_name)

        if SORT_HINT not in table:
            table[SORT_HINT] = get_columns_names_with_prop(table, "sort")  # type: ignore[typeddict-unknown-key]

        if PARTITION_HINT not in table:
            table[PARTITION_HINT] = get_columns_names_with_prop(table, "partition")  # type: ignore[typeddict-unknown-key]

        return table

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [ClickHouseMergeJob.from_table_chain(table_chain, self.sql_client)]

    def _get_column_def_sql(self, c: TColumnSchema, table: PreparedTableSchema = None) -> str:
        # Build column definition.
        # The primary key and sort order definition is defined outside column specification.
        hints_ = " ".join(
            self.active_hints.get(hint)
            for hint in self.active_hints.keys()
            if c.get(cast(str, hint), False) is True and hint not in ("primary_key", "sort")
        )

        if codec := c.get(CODEC_HINT):
            hints_ += f"CODEC({codec}) "

        # Alter table statements only accept `Nullable` modifiers.
        # JSON type isn't nullable in ClickHouse.
        type_with_nullability_modifier = (
            f"Nullable({self.type_mapper.to_destination_type(c,table)})"
            if is_nullable_column(c)
            else self.type_mapper.to_destination_type(c, table)
        )

        return (
            f"{self.sql_client.escape_column_name(c['name'])} {type_with_nullability_modifier} {hints_}"
            .strip()
        )

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        return super().create_load_job(table, file_path, load_id, restore) or ClickHouseLoadJob(
            file_path,
            config=self.config,
            staging_credentials=(
                self.config.staging_config.credentials if self.config.staging_config else None
            ),
        )

    def _get_key(
        self,
        table_schema: PreparedTableSchema,
        key_type: Literal["sort", "partition"],
    ) -> Optional[str]:
        """Returns sort or partition key based on hint."""

        hint = table_schema.get(SORT_HINT if key_type == "sort" else PARTITION_HINT)
        if not hint:
            return None

        assert isinstance(hint, (str, list, tuple))

        if isinstance(hint, str):
            # it's a SQL expression; return as is
            return hint
        elif isinstance(hint, (list, tuple)):
            # it's a sequence of column names; generate expression
            norm_hint_columns = [self.schema.naming.normalize_identifier(col) for col in hint]
            return "(" + ", ".join(norm_hint_columns) + ")"

    def _get_settings_clause(self, settings_hint: TMergeTreeSettings) -> str:
        def to_clickhouse_literal(val: TMergeTreeSettingsValue) -> str:
            if isinstance(val, str):
                escaped_str = self.capabilities.escape_literal(val)
                return cast(str, escaped_str)
            elif isinstance(val, bool):
                return "true" if val else "false"
            return str(val)

        clauses = [f"{key} = {to_clickhouse_literal(val)}" for key, val in settings_hint.items()]

        return ", ".join(clauses)

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        table = self.prepare_load_table(table_name)
        sql = SqlJobClientBase._get_table_update_sql(self, table_name, new_columns, generate_alter)

        if generate_alter:
            return sql

        # Default to 'MergeTree' if the user didn't explicitly set a table engine hint.
        # Clickhouse Cloud will automatically pick `SharedMergeTree` for this option,
        # so it will work on both local and cloud instances of CH.
        table_type = cast(
            TTableEngineType,
            table.get(
                cast(str, TABLE_ENGINE_TYPE_HINT),
                self.config.table_engine_type,
            ),
        )
        sql[0] = f"{sql[0]}\nENGINE = {TABLE_ENGINE_TYPE_TO_CLICKHOUSE_ATTR.get(table_type)}"

        # PRIMARY KEY
        if primary_key_list := [
            self.sql_client.escape_column_name(c["name"])
            for c in new_columns
            if c.get("primary_key")
        ]:
            sql[0] += "\nPRIMARY KEY (" + ", ".join(primary_key_list) + ")"
        else:
            sql[0] += "\nPRIMARY KEY tuple()"

        # ORDER BY
        if sort_key := self._get_key(table, "sort"):
            sql[0] += f"\nORDER BY {sort_key}"

        # PARTITION BY
        if part_key := self._get_key(table, "partition"):
            sql[0] += f"\nPARTITION BY {part_key}"

        # SETTNGS
        if settings_hint := table.get(SETTINGS_HINT):
            settings_hint = cast(TMergeTreeSettings, settings_hint)
            settings_clause = self._get_settings_clause(settings_hint)
            sql[0] += f"\nSETTINGS {settings_clause}"

        return sql

    def _gen_not_null(self, v: bool) -> str:
        # ClickHouse fields are not nullable by default.
        # We use the `Nullable` modifier instead of NULL / NOT NULL modifiers to cater for ALTER statement.
        return ""

    def _from_db_type(
        self, ch_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(ch_t, precision, scale)

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load



from typing import Any, Callable, Dict, Optional, TYPE_CHECKING, List
import socket
from dlt.common.typing import Annotated, TypeAlias, TSecretStrValue, SocketLike
from dlt.common.configuration.specs.base_configuration import (
    CredentialsConfiguration,
    configspec,
    NotResolved,
)

if TYPE_CHECKING:
    try:
        from paramiko import PKey
        from paramiko.auth_strategy import AuthStrategy
        from paramiko import Transport
    except ImportError:
        PKey = Any  # type: ignore[misc, assignment]
        AuthStrategy = Any  # type: ignore[misc, assignment]
        Transport = Any  # type: ignore[misc, assignment]
else:
    PKey = Any
    AuthStrategy = Any
    Transport = Any

SFTPTransportFactory: TypeAlias = Callable[[socket.socket], Transport]


@configspec
class SFTPCredentials(CredentialsConfiguration):
    """Credentials for SFTP filesystem, compatible with fsspec SFTP protocol.

    Authentication is attempted in the following order of priority:

        - `key_filename` may contain OpenSSH public certificate paths
          as well as regular private-key paths; when files ending in `-cert.pub` are found, they are assumed to match
          a private key, and both components will be loaded.

        - Any key found through an SSH agent: any “id_rsa”, “id_dsa”, or “id_ecdsa” key discoverable in ~/.ssh/.

        - Plain username/password authentication, if a password was provided.

        - If a private key requires a password to unlock it, and a password is provided, that password will be used to
          attempt to unlock the key.

    For more information about parameters:
    https://docs.paramiko.org/en/3.3/api/client.html#paramiko.client.SSHClient.connect
    """

    sftp_port: Optional[int] = 22
    sftp_username: Optional[str] = None
    sftp_password: Optional[TSecretStrValue] = None
    # Runtime-only pkey; cannot be loaded from env var, skip configspec.
    sftp_pkey: Annotated[Optional[PKey], NotResolved()] = None
    sftp_key_filename: Optional[str] = None
    sftp_key_passphrase: Optional[TSecretStrValue] = None
    sftp_timeout: Optional[float] = None
    sftp_banner_timeout: Optional[float] = None
    sftp_auth_timeout: Optional[float] = None
    sftp_channel_timeout: Optional[float] = None
    sftp_allow_agent: Optional[bool] = True
    sftp_look_for_keys: Optional[bool] = True
    sftp_compress: Optional[bool] = False
    # Runtime-only socket; cannot be loaded from env var, skip configspec.
    sftp_sock: Annotated[Optional[SocketLike], NotResolved()] = None
    sftp_gss_auth: Optional[bool] = False
    sftp_gss_kex: Optional[bool] = False
    sftp_gss_deleg_creds: Optional[bool] = True
    sftp_gss_host: Optional[str] = None
    sftp_gss_trust_dns: Optional[bool] = True
    # Runtime-only vars below; cannot be loaded from env var, skip configspec.
    sftp_disabled_algorithms: Annotated[Optional[Dict[str, List[str]]], NotResolved()] = None
    sftp_transport_factory: Annotated[Optional[SFTPTransportFactory], NotResolved()] = None
    sftp_auth_strategy: Annotated[Optional[AuthStrategy], NotResolved()] = None

    def to_fsspec_credentials(self) -> Dict[str, Any]:
        """Return a dict that can be passed to fsspec SFTP/SSHClient.connect method."""

        credentials: Dict[str, Any] = {
            "port": self.sftp_port,
            "username": self.sftp_username,
            "password": self.sftp_password,
            "pkey": self.sftp_pkey,
            "key_filename": self.sftp_key_filename,
            "passphrase": self.sftp_key_passphrase,
            "timeout": self.sftp_timeout,
            "banner_timeout": self.sftp_banner_timeout,
            "auth_timeout": self.sftp_auth_timeout,
            "channel_timeout": self.sftp_channel_timeout,
            "allow_agent": self.sftp_allow_agent,
            "look_for_keys": self.sftp_look_for_keys,
            "compress": self.sftp_compress,
            "sock": self.sftp_sock,
            "gss_auth": self.sftp_gss_auth,
            "gss_kex": self.sftp_gss_kex,
            "gss_deleg_creds": self.sftp_gss_deleg_creds,
            "gss_host": self.sftp_gss_host,
            "gss_trust_dns": self.sftp_gss_trust_dns,
            "disabled_algorithms": self.sftp_disabled_algorithms,
            "transport_factory": self.sftp_transport_factory,
            "auth_strategy": self.sftp_auth_strategy,
        }

        return credentials



"""
# DuckDB object hierarchy
Here are short definitions and relationships between DuckDB objects.
This should help disambiguate names used in Duckdb, DuckLake, and dlt.

TL;DR:
- scalar < column < table < schema (dataset) < database = catalog
- Typically, in duckdb, you have one catalog = one database = one file
- When using `ATTACH`, you're adding `Catalog` to your `Database`
    - Though if you do `SHOW ALL TABLES`, the result column "database" should be "catalog" to be precise

Hierarchy:
- A `Table` can have many `Column`
- A `Schema` can have many `Table`
- A `Database` can have many `Schema` (corresponds to dataset in dlt)
- A `Database` is a single physical file (e.g., `db.duckdb`)
- A `Database` has a single `Catalog`
- A `Catalog` is the internal metadata structure of everything found in the database
- Using `ATTACH` adds a `Catalog` to the

In dlt:
- dlt creates a duckdb `Database` per pipeline when using `dlt.pipeline(..., destination="duckdb")`
- dlt stores the data inside a `Schema` that matches the name of the `dlt.Dataset`
- when setting the pipeline destination to a specific duckdb `Database`, you can store multiple
`dlt.Dataset` inside the same instance (each with its own duckdb `Schema`).


# DuckLake object hierarchy

TL;DR:
- scalar < column < table < schema < snapshot < database = catalog

Hierarchy:
- A `Catalog` is an SQL database to store metadata
    - In duckdb terms, it's a duckdb `Database` that implements the duckdb `Catalog` for the DuckLake
- A `Catalog` has many Schemas (namespaces if you compare it to Iceberg) that correspond to dlt.Dataset
- A `Storage` is a file system or object store that can store parquet files
- A `Snapshot` references to the `Catalog` at a particular point in time
    - This places `Snapshot` at the top of the hierarchy because it scopes other constructs

Using the `ducklake` extension, the following command in duckdb

```sql
ATTACH 'ducklake:{catalog_database}' (DATA_PATH '{data_storage}');
```

adds the ducklake `Catalog` to your duckdb database
"""
from __future__ import annotations

import os
import pathlib
from packaging.version import Version
from typing import Any, Iterable, List, Optional, Sequence

import dlt
from dlt.common import logger
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import FollowupJobRequest, LoadJob
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.metrics import LoadJobMetrics
from dlt.common.schema.typing import TColumnSchema

from dlt.destinations.impl.duckdb.duck import DuckDbClient, DuckDbCopyJob
from dlt.destinations.impl.ducklake.sql_client import DuckLakeSqlClient
from dlt.destinations.impl.ducklake.configuration import DuckLakeClientConfiguration
from dlt.destinations.insert_job_client import InsertValuesJobClient
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.sql_client import SqlClientBase


class DuckLakeMergeFollowupJob(SqlMergeFollowupJob):
    """DuckLake-specific merge job that handles hard deletes separately.

    DuckLake doesn't support DELETE in MERGE statements, so we split into
    MERGE (upsert only) + separate DELETE for hard deletes.
    """

    @classmethod
    def gen_upsert_merge_sql(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        primary_keys: Sequence[str],
        root_table_column_names: Sequence[str],
        hard_delete_col: Optional[str],
        deleted_cond: Optional[str],
    ) -> List[str]:
        """Generate MERGE statement without DELETE clause + separate DELETE for hard deletes."""
        # Get MERGE statement from base class without DELETE clause (pass hard_delete_col=None)
        sql = SqlMergeFollowupJob.gen_upsert_merge_sql(
            root_table_name,
            staging_root_table_name,
            primary_keys,
            root_table_column_names,
            hard_delete_col=None,  # No DELETE in MERGE for DuckLake
            deleted_cond=None,
        )

        # Add separate DELETE for hard deletes
        if hard_delete_col is not None:
            sql.append(f"""
                DELETE FROM {root_table_name}
                WHERE ({", ".join(primary_keys)}) IN (
                    SELECT {", ".join(primary_keys)}
                    FROM {staging_root_table_name}
                    WHERE {deleted_cond}
                );
            """)
        return sql


class DuckLakeCopyJob(DuckDbCopyJob):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: DuckLakeClient = None

    def metrics(self) -> Optional[LoadJobMetrics]:
        """Generate remote url metrics which point to the table in storage"""
        m = super().metrics()
        # job client not available before run_managed called
        if not self._job_client:
            return m
        # TODO: read location from catalog. ducklake supports customized table layouts
        return m._replace(
            remote_url=str(
                pathlib.Path().joinpath(
                    self._job_client.config.credentials.storage.bucket_url,
                    self._job_client.sql_client.dataset_name,
                    self.load_table_name,
                )
            )
        )


class DuckLakeClient(DuckDbClient):
    """Destination client to interact with a DuckLake

    A DuckLake has 3 components:
        - ducklake client: this is a `duckdb` instance with the `ducklake` extension
        - catalog: this is an SQL database storing metadata. It can be a duckdb instance
            (typically the ducklake client) or a remote database (sqlite, postgres, mysql)
        - storage: this is a filesystem where data is stored in files

    The dlt DuckLake destination gives access to the "ducklake client".
    You never have to manage the catalog and storage directly;
    this is done through the ducklake client.
    """

    def __init__(
        self,
        schema: dlt.Schema,
        config: DuckLakeClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)  # type: ignore[arg-type]
        self.config: DuckLakeClientConfiguration = config  # type: ignore[assignment]
        self.sql_client: DuckLakeSqlClient = DuckLakeSqlClient(
            dataset_name=config.normalize_dataset_name(schema),
            staging_dataset_name=config.normalize_staging_dataset_name(schema),
            credentials=config.credentials,
            capabilities=capabilities,
            override_data_path=config.override_data_path,
        )

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        super().initialize_storage(truncate_tables)
        # create local storage
        if self.config.credentials.storage.is_local_filesystem:
            os.makedirs(self.config.credentials.storage_url, exist_ok=True)

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = InsertValuesJobClient.create_load_job(self, table, file_path, load_id, restore)
        if not job:
            job = DuckLakeCopyJob(file_path)
        return job

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        sql = super()._get_table_update_sql(table_name, new_columns, generate_alter)

        # TODO: evolve partitioning here. we should change the scheme if there's actual change vs.
        #  partition in schema
        if not generate_alter:
            partition_list = [
                self.sql_client.escape_column_name(c["name"])
                for c in new_columns
                if c.get("partition")
            ]

            if partition_list:
                import duckdb

                if Version(duckdb.__version__) < Version("1.4.0"):
                    logger.warning(
                        "Partitioning is disabled for duckdb < 1.4.0 due to malformed catalog"
                        " update statements."
                    )
                else:
                    qualified_name = self.sql_client.make_qualified_table_name(table_name)
                    sql.append(
                        f"ALTER TABLE {qualified_name} SET PARTITIONED BY ("
                        + ",".join(partition_list)
                        + ")"
                    )
        return sql

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [DuckLakeMergeFollowupJob.from_table_chain(table_chain, self.sql_client)]



import os
import tempfile

from dlt.common.configuration.specs.pluggable_run_context import (
    ProfilesRunContext,
)
from dlt.common.runtime.exceptions import RunContextNotAvailable
from dlt.common.runtime.run_context import is_folder_writable, switch_context

DEFAULT_WORKSPACE_WORKING_FOLDER = ".var"
DEFAULT_LOCAL_FOLDER = "_local"


def default_working_dir(run_dir: str, name: str, profile: str, default_folder: str) -> str:
    """Computes default data dir which is relative to `run_dir` and separated by `profile_name`
    If `project_dir` is not writable, we fall back to temp dir.
    """
    data_dir = os.path.join(run_dir, default_folder)
    if not is_folder_writable(run_dir):
        # fallback to temp dir which should be writable, project name is used to separate projects
        data_dir = os.path.join(tempfile.gettempdir(), "dlt", default_folder, name)
    return os.path.join(data_dir, profile)


def default_name(run_dir: str) -> str:
    run_dir = os.path.abspath(run_dir)
    name = os.path.basename(run_dir.rstrip(os.path.sep)) or "_dlt"
    return name


def switch_profile(profile: str) -> ProfilesRunContext:
    """Switch the active profile and reload the workspace run context.

    This is a convenience wrapper around `switch_context` that uses the
    `run_dir` of the current `active()` context.

    Args:
        profile: Name of the profile to activate.

    Returns:
        SupportsProfilesRunContext: The refreshed run context bound to the given profile.

    """
    from dlt.common.runtime.run_context import active

    ctx = active()
    # can switch profile only on a context that supports that
    context_class = ctx.__class__.__name__
    if not isinstance(ctx, ProfilesRunContext):
        raise RunContextNotAvailable(
            ctx.run_dir,
            f"Run context of type `{context_class}` located at `{ctx.run_dir}` does not support"
            " profiles.",
        )
    return switch_context(ctx.run_dir, profile=profile, required=context_class)  # type: ignore[return-value]



from typing import (
    List,
    Dict,
    Union,
    Literal,
    Callable,
    Any,
)
from requests import Response


HTTPMethodBasic = Literal["GET", "POST"]
HTTPMethodExtended = Literal["PUT", "PATCH", "DELETE", "HEAD", "OPTIONS"]
HTTPMethod = Union[HTTPMethodBasic, HTTPMethodExtended]
HookFunction = Callable[[Response, Any, Any], None]
HookEvent = Union[HookFunction, List[HookFunction]]
Hooks = Dict[str, HookEvent]



import random
from contextlib import contextmanager
from copy import copy
from types import TracebackType
from typing import (
    ClassVar,
    Dict,
    Iterator,
    Optional,
    Sequence,
    Type,
    Iterable,
    List,
)
import time
from dlt.common.metrics import LoadJobMetrics
from dlt.common.pendulum import pendulum
from dlt.common.schema import Schema, TSchemaTables
from dlt.common.storages import FileStorage
from dlt.common.storages.load_package import LoadJobInfo
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import (
    DestinationTerminalException,
    DestinationTransientException,
)
from dlt.common.destination.client import (
    HasFollowupJobs,
    FollowupJobRequest,
    PreparedTableSchema,
    SupportsStagingDestination,
    TLoadJobState,
    RunnableLoadJob,
    JobClientBase,
    WithStagingDataset,
    LoadJob,
)
from dlt.common.time import precise_time
from dlt.destinations.sql_jobs import SqlMergeFollowupJob

from dlt.destinations.exceptions import (
    LoadJobNotExistsException,
)
from dlt.destinations.impl.dummy.configuration import DummyClientConfiguration
from dlt.destinations.job_impl import ReferenceFollowupJobRequest


class LoadDummyBaseJob(RunnableLoadJob):
    def __init__(self, file_name: str, config: DummyClientConfiguration) -> None:
        super().__init__(file_name)
        self.config = copy(config)
        self.start_time = precise_time()

        if self.config.fail_terminally_in_init:
            raise DestinationTerminalException(self._exception)
        if self.config.fail_transiently_in_init:
            raise Exception(self._exception)

    def run(self) -> None:
        while True:
            # simulate generic exception (equals retry)
            c_r = random.random()
            if self.config.exception_prob >= c_r:
                # this will make the job go to a retry state with a generic exception
                raise Exception("Dummy job status raised exception")

            # timeout condition (terminal)
            n = precise_time()
            if n - self.start_time > self.config.timeout:
                # this will make the the job go to a failed state
                raise DestinationTerminalException("failed due to timeout")

            # success
            c_r = random.random()
            if self.config.completed_prob >= c_r:
                # this will make the run function exit and the job go to a completed state
                break

            # retry prob
            c_r = random.random()
            if self.config.retry_prob >= c_r:
                # this will make the job go to a retry state
                raise DestinationTransientException("a random retry occurred")

            # fail prob
            c_r = random.random()
            if self.config.fail_prob >= c_r:
                # this will make the the job go to a failed state
                raise DestinationTerminalException("a random fail occurred")

            time.sleep(0.1)

    def metrics(self) -> Optional[LoadJobMetrics]:
        m = super().metrics()
        # add remote url if there's followup job
        if self.config.create_followup_jobs:
            m = m._replace(remote_url=self._file_name)
        return m


class DummyFollowupJobRequest(ReferenceFollowupJobRequest):
    def __init__(
        self, original_file_name: str, remote_paths: List[str], config: DummyClientConfiguration
    ) -> None:
        self.config = config
        if config.fail_followup_job_creation:
            raise Exception("Failed to create followup job")
        super().__init__(original_file_name=original_file_name, remote_paths=remote_paths)


class LoadDummyJob(LoadDummyBaseJob, HasFollowupJobs):
    def create_followup_jobs(self, final_state: TLoadJobState) -> List[FollowupJobRequest]:
        if self.config.create_followup_jobs and final_state == "completed":
            new_job = DummyFollowupJobRequest(
                original_file_name=self.file_name(),
                remote_paths=[self._file_name],
                config=self.config,
            )
            CREATED_FOLLOWUP_JOBS[new_job.job_id()] = new_job
            return [new_job]
        return []


JOBS: Dict[str, LoadDummyBaseJob] = {}
CREATED_FOLLOWUP_JOBS: Dict[str, FollowupJobRequest] = {}
CREATED_TABLE_CHAIN_FOLLOWUP_JOBS: Dict[str, FollowupJobRequest] = {}
RETRIED_JOBS: Dict[str, LoadDummyBaseJob] = {}


class DummyClient(JobClientBase, SupportsStagingDestination, WithStagingDataset):
    """dummy client storing jobs in memory"""

    def __init__(
        self,
        schema: Schema,
        config: DummyClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        self.in_staging_context = False
        self.config: DummyClientConfiguration = config

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        pass

    def is_storage_initialized(self) -> bool:
        return True

    def drop_storage(self) -> None:
        pass

    def drop_tables(self, *tables: str, delete_schema: bool = True) -> None:
        pass

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        applied_update = super().update_stored_schema(only_tables, expected_update)
        if self.config.fail_schema_update:
            raise DestinationTransientException(
                "Raise on schema update due to `fail_schema_update` config flag"
            )
        return applied_update

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job_id = FileStorage.get_file_name_from_file_path(file_path)
        if restore and job_id not in JOBS:
            raise LoadJobNotExistsException(job_id)
        # return existing job if already there
        if job_id not in JOBS:
            JOBS[job_id] = self._create_job(file_path)
        else:
            job = JOBS[job_id]
            # update config of existing job in case it was changed in tests
            job.config = self.config
            RETRIED_JOBS[job_id] = job

        return JOBS[job_id]

    def create_table_chain_completed_followup_jobs(
        self,
        table_chain: Sequence[PreparedTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]] = None,
    ) -> List[FollowupJobRequest]:
        """Creates a list of followup jobs that should be executed after a table chain is completed"""

        # if sql job follow up is configure we schedule a merge job that will always fail
        if self.config.fail_table_chain_followup_job_creation:
            raise Exception("Failed to create table chain followup job")
        if self.config.create_followup_table_chain_sql_jobs:
            return [SqlMergeFollowupJob.from_table_chain(table_chain, self)]  # type: ignore
        if self.config.create_followup_table_chain_reference_jobs:
            table_job_paths = [job.file_path for job in completed_table_chain_jobs]
            file_name = FileStorage.get_file_name_from_file_path(table_job_paths[0])
            job = ReferenceFollowupJobRequest(file_name, table_job_paths)
            CREATED_TABLE_CHAIN_FOLLOWUP_JOBS[job.job_id()] = job
            return [job]
        return []

    def complete_load(self, load_id: str) -> None:
        pass

    def should_load_data_to_staging_dataset(self, table_name: str) -> bool:
        return super().should_load_data_to_staging_dataset(table_name)

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load

    @contextmanager
    def with_staging_dataset(self) -> Iterator[JobClientBase]:
        try:
            self.in_staging_context = True
            yield self
        finally:
            self.in_staging_context = False

    def __enter__(self) -> "DummyClient":
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        pass

    def _create_job(self, job_id: str) -> LoadDummyBaseJob:
        if ReferenceFollowupJobRequest.is_reference_job(job_id):
            return LoadDummyBaseJob(job_id, config=self.config)
        else:
            return LoadDummyJob(job_id, config=self.config)



import os
from typing import Dict, Optional, Sequence, List, cast, Union
from urllib.parse import urlparse
from pathlib import Path

from dlt.common.configuration.specs.azure_credentials import (
    AzureServicePrincipalCredentialsWithoutDefaults,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    HasFollowupJobs,
    FollowupJobRequest,
    PreparedTableSchema,
    RunnableLoadJob,
    SupportsStagingDestination,
    LoadJob,
)
from dlt.common.configuration.specs import (
    AwsCredentialsWithoutDefaults,
    AzureCredentialsWithoutDefaults,
)
from dlt.common.schema.utils import get_columns_names_with_prop
from dlt.common.storages.configuration import ensure_canonical_az_url
from dlt.common.storages.file_storage import FileStorage
from dlt.common.storages.fsspec_filesystem import (
    AZURE_BLOB_STORAGE_PROTOCOLS,
    S3_PROTOCOLS,
    GCS_PROTOCOLS,
)
from dlt.destinations.impl.databricks.databricks_adapter import (
    CLUSTER_HINT,
    TABLE_PROPERTIES_HINT,
    TABLE_COMMENT_HINT,
    TABLE_TAGS_HINT,
    COLUMN_COMMENT_HINT,
    COLUMN_TAGS_HINT,
)
from dlt.common.schema import TColumnSchema, Schema, TTableSchema, TColumnHint
from dlt.common.schema.typing import TColumnType
from dlt.common.storages import FilesystemConfiguration, fsspec_from_config
from dlt.common.utils import uniq_id
from dlt.common import logger
from dlt.common.data_writers.escape import escape_databricks_literal
from dlt.common.exceptions import TerminalValueError
from dlt.destinations.job_client_impl import SqlJobClientWithStagingDataset
from dlt.destinations.exceptions import LoadJobTerminalException
from dlt.destinations.impl.databricks.configuration import DatabricksClientConfiguration
from dlt.destinations.impl.databricks.sql_client import DatabricksSqlClient
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.impl.databricks.typing import TDatabricksColumnHint
from dlt.destinations.path_utils import get_file_format_and_compression

SUPPORTED_BLOB_STORAGE_PROTOCOLS = AZURE_BLOB_STORAGE_PROTOCOLS + S3_PROTOCOLS + GCS_PROTOCOLS


class DatabricksLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(
        self,
        file_path: str,
        staging_config: FilesystemConfiguration,
    ) -> None:
        super().__init__(file_path)
        self._staging_config = staging_config
        self._job_client: "DatabricksClient" = None

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client

        qualified_table_name = self._sql_client.make_qualified_table_name(self.load_table_name)

        # decide if this is a local file or a staged file
        is_local_file = not ReferenceFollowupJobRequest.is_reference_job(self._file_path)
        if is_local_file:
            # staging_allowed_local_path should be set when opening the connection but at that
            # time we do not know this path so do it now
            conn_ = self._sql_client.native_connection
            file_dir = os.path.dirname(self._file_path)
            if backend := getattr(conn_, "thrift_backend", None):
                backend.staging_allowed_local_path = file_dir
            else:
                # thrift backend discontinued on newer databricks connector clients
                conn_.staging_allowed_local_path = file_dir  # type: ignore[attr-defined,unused-ignore]
            # local file by uploading to a temporary volume on Databricks
            from_clause, file_name, volume_path, volume_file_path = self._handle_local_file_upload(
                self._file_path
            )
            credentials_clause = ""
            orig_bucket_path = None  # not used for local file
        else:
            # staged file
            from_clause, credentials_clause, file_name, orig_bucket_path = (
                self._handle_staged_file()
            )

        # decide on source format, file_name will either be a local file or a bucket path
        source_format, format_options_clause, skip_load = self._determine_source_format(
            file_name, orig_bucket_path
        )

        if skip_load:
            # If the file is empty or otherwise un-loadable, exit early
            return

        statement = self._build_copy_into_statement(
            qualified_table_name,
            from_clause,
            credentials_clause,
            source_format,
            format_options_clause,
        )

        self._sql_client.execute_sql(statement)

        if is_local_file and not self._job_client.config.keep_staged_files:
            self._handle_staged_file_remove(volume_path, volume_file_path)

    def _handle_staged_file_remove(self, volume_path: str, volume_file_path: str) -> None:
        self._sql_client.execute_sql(f"REMOVE '{volume_file_path}'")
        self._sql_client.execute_sql(f"REMOVE '{volume_path}'")

    def _handle_local_file_upload(self, local_file_path: str) -> tuple[str, str, str, str]:
        file_name = FileStorage.get_file_name_from_file_path(local_file_path)
        volume_file_name = file_name
        if file_name.startswith(("_", ".")):
            volume_file_name = (
                "valid" + file_name
            )  # databricks loading fails when file_name starts with - or .

        volume_catalog = self._sql_client.database_name
        volume_database = self._sql_client.dataset_name
        volume_name = "_dlt_staging_load_volume"

        fully_qualified_volume_name = f"{volume_catalog}.{volume_database}.{volume_name}"
        if self._job_client.config.staging_volume_name:
            fully_qualified_volume_name = self._job_client.config.staging_volume_name
            volume_catalog, volume_database, volume_name = fully_qualified_volume_name.split(".")
        else:
            # create staging volume named _dlt_staging_load_volume
            self._sql_client.execute_sql(f"""
                CREATE VOLUME IF NOT EXISTS {fully_qualified_volume_name}
            """)

        volume_path = f"/Volumes/{volume_catalog}/{volume_database}/{volume_name}/{uniq_id()}"
        volume_file_path = f"{volume_path}/{volume_file_name}"

        posix_path = Path(
            local_file_path
        ).as_posix()  # backslash in Windows local path causes issues with PUT command
        self._sql_client.execute_sql(f"PUT '{posix_path}' INTO '{volume_file_path}' OVERWRITE")

        from_clause = f"FROM '{volume_path}'"

        return from_clause, file_name, volume_path, volume_file_path

    def _handle_staged_file(self) -> tuple[str, str, str, str]:
        bucket_path = orig_bucket_path = (
            ReferenceFollowupJobRequest.resolve_reference(self._file_path)
            if ReferenceFollowupJobRequest.is_reference_job(self._file_path)
            else ""
        )

        if not bucket_path:
            raise LoadJobTerminalException(
                self._file_path,
                "Cannot load from local file. Databricks does not support loading from local files."
                " Configure staging with an s3, azure or google storage bucket.",
            )

        file_name = FileStorage.get_file_name_from_file_path(bucket_path)

        staging_credentials = self._staging_config.credentials
        bucket_url = urlparse(bucket_path)
        bucket_scheme = bucket_url.scheme

        if bucket_scheme not in SUPPORTED_BLOB_STORAGE_PROTOCOLS:
            raise LoadJobTerminalException(
                self._file_path,
                f"Databricks cannot load data from staging bucket `{bucket_path}`. "
                "Only s3, azure and gcs buckets are supported. "
                "Please note that gcs buckets are supported only via named credential.",
            )

        credentials_clause = ""

        if self._job_client.config.is_staging_external_location:
            # just skip the credentials clause for external location
            # https://docs.databricks.com/en/sql/language-manual/sql-ref-external-locations.html#external-location
            pass
        elif self._job_client.config.staging_credentials_name:
            # add named credentials
            credentials_clause = (
                f"WITH(CREDENTIAL {self._job_client.config.staging_credentials_name} )"
            )
        else:
            # referencing an staged files via a bucket URL requires explicit AWS credentials
            if bucket_scheme == "s3":
                assert isinstance(staging_credentials, AwsCredentialsWithoutDefaults)
                s3_creds = staging_credentials.to_session_credentials()
                credentials_clause = f"""WITH(CREDENTIAL(
                    AWS_ACCESS_KEY='{s3_creds["aws_access_key_id"]}',
                    AWS_SECRET_KEY='{s3_creds["aws_secret_access_key"]}',
                    AWS_SESSION_TOKEN='{s3_creds["aws_session_token"]}'
                ))"""
            elif bucket_scheme in AZURE_BLOB_STORAGE_PROTOCOLS:
                assert isinstance(
                    staging_credentials, AzureCredentialsWithoutDefaults
                ), "AzureCredentialsWithoutDefaults required to pass explicit credential"
                # Explicit azure credentials are needed to load from bucket without a named stage
                credentials_clause = f"""WITH(CREDENTIAL(AZURE_SAS_TOKEN='{staging_credentials.azure_storage_sas_token}'))"""
                bucket_path = self.ensure_databricks_abfss_url(
                    bucket_path,
                    staging_credentials.azure_storage_account_name,
                    staging_credentials.azure_account_host,
                )
            else:
                raise LoadJobTerminalException(
                    self._file_path,
                    "You need to use Databricks named credential to use google storage."
                    " Passing explicit Google credentials is not supported by Databricks.",
                )

        if bucket_scheme in AZURE_BLOB_STORAGE_PROTOCOLS:
            assert isinstance(
                staging_credentials,
                (AzureCredentialsWithoutDefaults, AzureServicePrincipalCredentialsWithoutDefaults),
            )
            bucket_path = self.ensure_databricks_abfss_url(
                bucket_path,
                staging_credentials.azure_storage_account_name,
                staging_credentials.azure_account_host,
            )

        # always add FROM clause
        from_clause = f"FROM '{bucket_path}'"

        return from_clause, credentials_clause, file_name, orig_bucket_path

    def _determine_source_format(
        self, file_name: str, orig_bucket_path: str
    ) -> tuple[str, str, bool]:
        file_format, _ = get_file_format_and_compression(file_name)

        if file_format == "parquet":
            return "PARQUET", "", False

        elif file_format in ["jsonl", "typed-jsonl"]:
            format_options_clause = "FORMAT_OPTIONS('inferTimestamp'='true')"

            # check for an empty JSON file, unless it's a direct load
            if self._staging_config is not None:
                fs, _ = fsspec_from_config(self._staging_config)
                if orig_bucket_path is not None:
                    file_size = fs.size(orig_bucket_path)
                    if file_size == 0:
                        return "JSON", format_options_clause, True

            return "JSON", format_options_clause, False

        raise LoadJobTerminalException(
            self._file_path, "Databricks loader only supports .parquet or .jsonl file extensions."
        )

    def _build_copy_into_statement(
        self,
        qualified_table_name: str,
        from_clause: str,
        credentials_clause: str,
        source_format: str,
        format_options_clause: str,
    ) -> str:
        return f"""COPY INTO {qualified_table_name}
            {from_clause}
            {credentials_clause}
            FILEFORMAT = {source_format}
            {format_options_clause}
        """

    @staticmethod
    def ensure_databricks_abfss_url(
        bucket_path: str, azure_storage_account_name: str = None, account_host: str = None
    ) -> str:
        return ensure_canonical_az_url(
            bucket_path, "abfss", azure_storage_account_name, account_host
        )


class DatabricksMergeJob(SqlMergeFollowupJob):
    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        return f"CREATE TEMPORARY VIEW {temp_table_name} AS {select_sql}"

    @classmethod
    def gen_delete_from_sql(
        cls, table_name: str, column_name: str, temp_table_name: str, temp_table_column: str
    ) -> str:
        # Databricks does not support subqueries in DELETE FROM statements so we use a MERGE statement instead
        return f"""MERGE INTO {table_name}
        USING {temp_table_name}
        ON {table_name}.{column_name} = {temp_table_name}.{temp_table_column}
        WHEN MATCHED THEN DELETE
        """


class DatabricksClient(SqlJobClientWithStagingDataset, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: DatabricksClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = SqlJobClientWithStagingDataset.create_dataset_names(
            schema, config
        )
        sql_client = DatabricksSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.config: DatabricksClientConfiguration = config
        self.sql_client: DatabricksSqlClient = sql_client  # type: ignore[assignment, unused-ignore]
        self.type_mapper = self.capabilities.get_type_mapper()
        # PK and FK are created in SQL fragments, not inline

    def _get_column_def_sql(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        column_def_sql = super()._get_column_def_sql(column, table)

        if column.get(COLUMN_COMMENT_HINT) or column.get("description"):
            comment = column.get(COLUMN_COMMENT_HINT) or column.get("description")
            escaped_comment = escape_databricks_literal(comment)
            column_def_sql = f"{column_def_sql} COMMENT {escaped_comment}"
        return column_def_sql

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)

        if not job:
            job = DatabricksLoadJob(
                file_path,
                staging_config=cast(FilesystemConfiguration, self.config.staging_config),
            )
        return job

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [DatabricksMergeJob.from_table_chain(table_chain, self.sql_client)]

    def _make_add_column_sql(
        self, new_columns: Sequence[TColumnSchema], table: PreparedTableSchema = None
    ) -> List[str]:
        # Override because databricks requires multiple columns in a single ADD COLUMN clause
        return [
            "ADD COLUMN\n" + ",\n".join(self._get_column_def_sql(c, table) for c in new_columns)
        ]

    def _get_constraints_sql(
        self,
        table_name: str,
        new_columns: Sequence[TColumnSchema],
        generate_alter: bool,
    ) -> str:
        constraints_sql = ""

        partial: TTableSchema = {
            "name": table_name,
            "columns": {c["name"]: c for c in new_columns},
        }

        if self.config.create_indexes:
            # Adding primary key constraint
            pk_columns = get_columns_names_with_prop(partial, "primary_key")

            if pk_columns:
                logger.info(f"Creating PRIMARY KEY constraint for table {table_name}")
                if generate_alter:
                    logger.warning(
                        f"PRIMARY KEY on {table_name} constraint cannot be added in ALTER TABLE and"
                        " is ignored"
                    )
                else:
                    quoted_pk_cols = ", ".join(
                        self.sql_client.escape_column_name(col) for col in pk_columns
                    )
                    constraints_sql += f",\nPRIMARY KEY ({quoted_pk_cols})"

            return constraints_sql

        return ""

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        table = self.prepare_load_table(table_name)

        # Check for AUTO clustering at table level
        cluster_by_auto = table.get(CLUSTER_HINT) == "AUTO"

        # Get cluster columns from column hints
        cluster_list = [
            self.sql_client.escape_column_name(c["name"])
            for c in new_columns
            if c.get("cluster") or c.get(CLUSTER_HINT, False)
        ]

        # Get partition columns from column hints
        partition_list = [
            self.sql_client.escape_column_name(c["name"])
            for c in new_columns
            if c.get("partition", False)
        ]

        # Determine the CLUSTER BY clause
        cluster_clause = None
        if cluster_by_auto:
            cluster_clause = "CLUSTER BY AUTO"
        elif cluster_list:
            cluster_clause = f"CLUSTER BY ({','.join(cluster_list)})"

        # Determine the PARTITIONED BY clause
        partition_clause = None
        if partition_list:
            partition_clause = f"PARTITIONED BY ({','.join(partition_list)})"

        # Get table properties
        table_properties = table.get(TABLE_PROPERTIES_HINT)
        tblproperties_clause = None
        if table_properties and isinstance(table_properties, dict):
            props = []
            for key, value in table_properties.items():
                # Escape key and value properly
                escaped_key = f"'{key}'"
                if isinstance(value, str):
                    escaped_value = f"'{value}'"
                elif isinstance(value, bool):
                    escaped_value = str(value).lower()
                else:
                    escaped_value = str(value)
                props.append(f"{escaped_key}={escaped_value}")
            tblproperties_clause = f"TBLPROPERTIES ({', '.join(props)})"

        # Get table format
        table_format = table.get("table_format", "delta")
        using_clause = None
        if table_format == "iceberg":
            using_clause = "USING ICEBERG"

            # Validate Iceberg-specific constraints
            if table_properties and isinstance(table_properties, dict):
                # Check for Delta-specific properties that are not supported in Iceberg
                delta_only_props = [
                    "delta.dataSkippingStatsColumns",
                    "delta.autoOptimize.optimizeWrite",
                    "delta.autoOptimize.autoCompact",
                    "delta.logRetentionDuration",
                    "delta.deletedFileRetentionDuration",
                    "delta.enableChangeDataFeed",
                    "delta.columnMapping.mode",
                    "delta.appendOnly",
                ]

                for prop_key in table_properties.keys():
                    if any(
                        prop_key.startswith(delta_prop) or prop_key == delta_prop
                        for delta_prop in delta_only_props
                    ):
                        raise TerminalValueError(
                            f"Table property '{prop_key}' is Delta Lake specific and not supported"
                            " with ICEBERG tables. Remove this property when using"
                            " table_format='iceberg'."
                        )
        # Note: DELTA is the default format, no explicit USING clause needed

        # For CREATE TABLE, we need custom generation if we have any custom clauses or non-DELTA format
        if not generate_alter and (
            cluster_clause or partition_clause or tblproperties_clause or using_clause
        ):
            # Build CREATE TABLE with all custom clauses
            qualified_name = self.sql_client.make_qualified_table_name(table_name)
            sql = self._make_create_table(qualified_name, table)

            sql += " (\n"
            sql += ",\n".join([self._get_column_def_sql(c, table) for c in new_columns])
            sql += self._get_constraints_sql(table_name, new_columns, generate_alter)
            sql += ")"

            # Add USING clause after column definitions for ICEBERG
            if using_clause:
                sql += f" {using_clause}"

            # Add PARTITIONED BY clause (must come before CLUSTER BY)
            if partition_clause:
                sql += f" {partition_clause}"
            # Add CLUSTER BY clause
            if cluster_clause:
                sql += f" {cluster_clause}"
            # Add TBLPROPERTIES clause (comes after CLUSTER BY)
            if tblproperties_clause:
                sql += f" {tblproperties_clause}"
            sql_result = [sql]
        else:
            # Use parent implementation for ALTER or non-clustered/non-partitioned tables
            sql_result = super()._get_table_update_sql(table_name, new_columns, generate_alter)

            # For ALTER TABLE, add CLUSTER BY as a separate statement
            # Note: PARTITIONED BY cannot be added via ALTER TABLE in Databricks
            if generate_alter and cluster_clause:
                qualified_name = self.sql_client.make_qualified_table_name(table_name)
                sql_result.append(f"ALTER TABLE {qualified_name} {cluster_clause}")

        qualified_name = self.sql_client.make_qualified_table_name(table_name)

        if table.get(TABLE_COMMENT_HINT) or table.get("description"):
            comment = table.get(TABLE_COMMENT_HINT) or table.get("description")
            escaped_comment = escape_databricks_literal(comment)
            sql_result.append(f"COMMENT ON TABLE {qualified_name} IS {escaped_comment}")

        if table.get(TABLE_TAGS_HINT):
            table_tags = cast(List[Union[str, Dict[str, str]]], table.get(TABLE_TAGS_HINT))
            for tag in table_tags:
                if isinstance(tag, str):
                    escaped_tag = escape_databricks_literal(tag)
                    sql_result.append(f"ALTER TABLE {qualified_name} SET TAGS ({escaped_tag})")
                elif isinstance(tag, dict):
                    (key, value), *rest = tag.items()
                    escaped_key = escape_databricks_literal(key)
                    escaped_value = escape_databricks_literal(value)
                    sql_result.append(
                        f"ALTER TABLE {qualified_name} SET TAGS ({escaped_key}={escaped_value})"
                    )

        column_tag_list = [
            (self.sql_client.escape_column_name(c["name"]), c.get(COLUMN_TAGS_HINT))
            for c in new_columns
            if c.get(COLUMN_TAGS_HINT)
        ]

        if column_tag_list:
            for column_name, column_tags in column_tag_list:
                column_tags_typed = cast(List[Union[str, Dict[str, str]]], column_tags)
                for column_tag in column_tags_typed:
                    if isinstance(column_tag, str):
                        escaped_tag = escape_databricks_literal(column_tag)
                        sql_result.append(
                            f"ALTER TABLE {qualified_name} ALTER COLUMN {column_name} SET TAGS"
                            f" ({escaped_tag})"
                        )
                    elif isinstance(column_tag, dict):
                        (key, value), *rest = column_tag.items()
                        escaped_key = escape_databricks_literal(key)
                        escaped_value = escape_databricks_literal(value)
                        sql_result.append(
                            f"ALTER TABLE {qualified_name} ALTER COLUMN {column_name} SET TAGS"
                            f" ({escaped_key}={escaped_value})"
                        )

        return sql_result

    def _get_table_post_update_sql(
        self,
        table: TTableSchema,
    ) -> List[str]:
        sql: List[str] = []
        table_name = table["name"]

        # add foreign key constraints
        references = table.get("references")
        if references:
            logger.info(f"Creating FOREIGN KEY constraint for table {table_name}")
            for reference in references:
                quoted_fk_cols = ", ".join(
                    self.sql_client.escape_column_name(col) for col in reference.get("columns")
                )
                quoted_reference_cols = ", ".join(
                    self.sql_client.escape_column_name(col)
                    for col in reference.get("referenced_columns")
                )
                sql.append(
                    f"ALTER TABLE {self.sql_client.make_qualified_table_name(table_name)} ADD"
                    f" FOREIGN KEY ({quoted_fk_cols}) REFERENCES"
                    f" {self.sql_client.make_qualified_table_name(reference.get('referenced_table'))}({quoted_reference_cols})"
                )

        return sql

    def _from_db_type(
        self, bq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(bq_t, precision, scale)

    def _get_storage_table_query_columns(self) -> List[str]:
        fields = super()._get_storage_table_query_columns()
        fields[2] = (  # Override because this is the only way to get data type with precision
            "full_data_type"
        )
        return fields

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load



"""Implements SupportsTracking"""
from typing import Any, ClassVar, List, Optional, Tuple, Union
import fsspec
import pickle
import os

import dlt
from dlt.common import logger
from dlt.common.configuration.exceptions import ConfigurationException
from dlt.common.configuration.resolve import resolve_configuration
from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec
from dlt.common.storages.configuration import FilesystemConfiguration
from dlt.common.storages.fsspec_filesystem import FileItemDict, fsspec_from_config, glob_files
from dlt.common.versioned_state import json_encode_state

from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace, TPipelineStep, SupportsPipeline
from dlt._workspace.run_context import DEFAULT_WORKSPACE_WORKING_FOLDER
from dlt._workspace._workspace_context import WorkspaceRunContext


@configspec
class RuntimeArtifactsConfiguration(BaseConfiguration):
    artifacts: FilesystemConfiguration = None


def sync_from_runtime() -> None:
    """Sync the pipeline states and traces from the runtime backup, recursively."""
    from dlt._workspace.helpers.runtime.runtime_artifacts import _get_runtime_artifacts_fs

    def sync_dir(fs: fsspec.AbstractFileSystem, src_root: str, dst_root: str) -> None:
        """Recursively sync src_root on fs into dst_root locally, always using fs.walk."""

        os.makedirs(dst_root, exist_ok=True)

        for file_dict in glob_files(fs, src_root):
            file_item = FileItemDict(file_dict, fs)

            relative_dir = os.path.dirname(file_dict["relative_path"])
            local_dir = dst_root if relative_dir == "." else os.path.join(dst_root, relative_dir)
            os.makedirs(local_dir, exist_ok=True)

            local_file = os.path.join(dst_root, file_dict["relative_path"])

            logger.info(f"Restoring artifact {local_file}")
            with open(local_file, "wb") as lf:
                lf.write(file_item.read_bytes())

            ts = file_dict["modification_date"].timestamp()
            os.utime(local_file, (ts, ts))  # (atime, mtime)

    context = dlt.current.run_context()

    if not context.runtime_config.run_id:
        return

    if not isinstance(context, WorkspaceRunContext):
        return

    fs, config = _get_runtime_artifacts_fs(section="sync")
    if not fs:
        return

    # TODO: there's no good way to get this value on sync.
    data_dir_root = os.path.join(
        context.settings_dir, DEFAULT_WORKSPACE_WORKING_FOLDER
    )  # the local .var folder

    # Just sync the whole base folder into the local pipelines dir
    sync_dir(fs, config.bucket_url, data_dir_root)


def _get_runtime_artifacts_fs(
    section: str,
) -> Tuple[fsspec.AbstractFileSystem, FilesystemConfiguration]:
    try:
        config = resolve_configuration(RuntimeArtifactsConfiguration(), sections=(section,))
    except ConfigurationException:
        logger.info(f"No artifact storage credentials found for {section}")
        return None, None

    return fsspec_from_config(config.artifacts)[0], config.artifacts


def _write_to_bucket(
    fs: fsspec.AbstractFileSystem,
    bucket_url: str,
    pipeline_name: str,
    paths: List[str],
    data: Union[str, bytes],
) -> None:
    # write to bucket using the config, same object may be written to multiple paths

    logger.info(f"Will send run artifact to {bucket_url}: {paths}")
    for path in paths:
        with fs.open(f"{bucket_url}/{pipeline_name}/{path}", mode="wb") as f:
            # encode text data as utf-8
            if isinstance(data, str):
                data = data.encode("utf-8")
            f.write(data)


def _send_trace_to_bucket(
    fs: fsspec.AbstractFileSystem, bucket_url: str, trace: PipelineTrace, pipeline: SupportsPipeline
) -> None:
    """
    Send the full trace pickled to the runtime bucket
    """
    pickled_trace = pickle.dumps(trace)
    _write_to_bucket(
        fs,
        bucket_url,
        pipeline.pipeline_name,
        [
            "trace.pickle",
        ],  # save current and by start time
        pickled_trace,
    )


def _send_state_to_bucket(
    fs: fsspec.AbstractFileSystem, bucket_url: str, pipeline: SupportsPipeline
) -> None:
    encoded_state = json_encode_state(pipeline.state)
    _write_to_bucket(
        fs,
        bucket_url,
        pipeline.pipeline_name,
        [
            "state.json",
        ],  # save current and by start time
        encoded_state,
    )


def _send_schemas_to_bucket(
    fs: fsspec.AbstractFileSystem, bucket_url: str, pipeline: SupportsPipeline
) -> None:
    schema_dir = os.path.join(pipeline.working_dir, "schemas")
    for schema_file in os.listdir(schema_dir):
        _write_to_bucket(
            fs,
            bucket_url,
            pipeline.pipeline_name,
            [f"schemas/{schema_file}"],
            open(os.path.join(schema_dir, schema_file), "rb").read(),
        )


def on_start_trace(trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline) -> None:
    pass


def on_start_trace_step(
    trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
) -> None:
    pass


def on_end_trace_step(
    trace: PipelineTrace,
    step: PipelineStepTrace,
    pipeline: SupportsPipeline,
    step_info: Any,
    send_state: bool,
) -> None:
    pass


def on_end_trace(trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool) -> None:
    # skip if runtime not running
    if pipeline.run_context.runtime_config.run_id is None:
        return

    fs, config = _get_runtime_artifacts_fs(section="send")
    if fs:
        logger.info(
            f"Sending run artifacts from pipeline `{pipeline.pipeline_name}` to"
            f" `{config.bucket_url}`"
        )
        try:
            _send_trace_to_bucket(fs, config.bucket_url, trace, pipeline)
            _send_state_to_bucket(fs, config.bucket_url, pipeline)
            _send_schemas_to_bucket(fs, config.bucket_url, pipeline)
        except Exception:
            logger.exception(
                f"Sending run artifacts from pipeline `{pipeline.pipeline_name}` to"
                f" `{config.bucket_url}`"
            )
            raise
        else:
            logger.info("Pipeline results reported to runtime")



from typing import Any, Sequence

from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.configuration.container import Container
from dlt.common.destination import DestinationCapabilitiesContext


def row_tuples_to_arrow(
    rows: Sequence[Any],
    columns: TTableSchemaColumns = None,
    tz: str = None,
) -> Any:
    """Converts `column_schema` to arrow schema using `caps` and `tz`. `caps` are injected from the container - which
    is always the case if run within the pipeline. This will generate arrow schema compatible with the destination.
    Otherwise generic capabilities are used
    """
    from dlt.common.libs.pyarrow import row_tuples_to_arrow as _row_tuples_to_arrow

    return _row_tuples_to_arrow(
        rows=rows,
        caps=Container().get(DestinationCapabilitiesContext)
        or DestinationCapabilitiesContext.generic_capabilities(),
        columns=columns,
        tz=tz,
        safe_arrow_conversion=True,
    )



from typing import Any, Dict, List, Optional

from pydantic import Field
from fastmcp.exceptions import ToolError

from dlt.common.typing import Annotated
from dlt._workspace.cli.ai.utils import (
    fetch_secrets_list,
    fetch_secrets_update_fragment,
    fetch_secrets_view_redacted,
)
from dlt._workspace.mcp.context import with_mcp_tool_telemetry


@with_mcp_tool_telemetry()
def secrets_list() -> List[Dict[str, Any]]:
    """List secret file paths, profiles, and whether each file exists."""

    return fetch_secrets_list()  # type: ignore[return-value]


@with_mcp_tool_telemetry()
def secrets_view_redacted(
    path: Annotated[
        Optional[str],
        Field(
            description=(
                "Absolute path to a specific secrets file (from secrets_list)."
                " Omit to get the unified merged view of all secret files."
            )
        ),
    ] = None,
) -> str:
    """Show secrets TOML with every value replaced by '***'.

    Without path: returns the unified merged view across all project secret files
    (read-only, shows effective configuration). With path: returns that single file
    redacted. Use secrets_list to discover file paths.
    """
    result = fetch_secrets_view_redacted(path)
    if result is None:
        if path:
            raise ToolError("Secrets file not found: %s" % path)
        raise ToolError("No secrets found in project providers.")
    return result


@with_mcp_tool_telemetry()
def secrets_update_fragment(
    fragment: Annotated[
        str,
        Field(description="TOML fragment to deep-merge into the secrets file"),
    ],
    path: Annotated[
        str,
        Field(description="Absolute path to the secrets file to update (from secrets_list)."),
    ],
) -> str:
    """Deep-merge a TOML fragment into a secrets file; returns the redacted result. The file is created if it does not exist."""
    try:
        return fetch_secrets_update_fragment(fragment, path)
    except Exception as ex:
        raise ToolError("Invalid TOML fragment: %s" % str(ex)) from ex


__tools__ = (secrets_list, secrets_view_redacted, secrets_update_fragment)



from collections.abc import Sequence as C_Sequence
from typing import Any, Dict, Sequence, Union

from dlt.common.typing import TDataItem
from dlt.extract.items_transform import ItemTransformFunctionNoMeta

import jsonpath_ng


def take_first(max_items: int) -> ItemTransformFunctionNoMeta[bool]:
    """A filter that takes only first `max_items` from a resource"""
    count: int = 0

    def _filter(_: TDataItem) -> bool:
        nonlocal count
        count += 1
        return count <= max_items

    return _filter


def skip_first(max_items: int) -> ItemTransformFunctionNoMeta[bool]:
    """A filter that skips first `max_items` from a resource"""
    count: int = 0

    def _filter(_: TDataItem) -> bool:
        nonlocal count
        count += 1
        return count > max_items

    return _filter


def pivot(
    paths: Union[str, Sequence[str]] = "$", prefix: str = "col"
) -> ItemTransformFunctionNoMeta[TDataItem]:
    """
    Pivot the given sequence of sequences into a sequence of dicts,
    generating column names from the given prefix and indexes, e.g.:
    {"field": [[1, 2]]} -> {"field": [{"prefix_0": 1, "prefix_1": 2}]}

    Args:
        paths (Union[str, Sequence[str]]): JSON paths of the fields to pivot.
        prefix (Optional[str]): Prefix to add to the column names.

    Returns:
        ItemTransformFunctionNoMeta[TDataItem]: The transformer function.
    """
    if isinstance(paths, str):
        paths = [paths]

    def _seq_to_dict(seq: Sequence[Any]) -> Dict[str, Any]:
        """
        Transform the given sequence into a dict, generating
        columns with the given prefix.

        Args:
            seq (List): The sequence to transform.

        Returns:
            Dict: a dictionary with the sequence values.
        """
        return {prefix + str(i): value for i, value in enumerate(seq)}

    def _raise_if_not_sequence(match: jsonpath_ng.jsonpath.DatumInContext) -> None:
        """Check if the given field is a sequence of sequences.

        Args:
            match (jsonpath_ng.jsonpath.DatumInContext): The field to check.
        """
        if not isinstance(match.value, C_Sequence):
            raise ValueError(
                "Pivot transformer is only applicable to sequences "
                f"fields, however, the value of `{str(match.full_path)}`"
                " is not a sequence."
            )

        for item in match.value:
            if not isinstance(item, C_Sequence):
                raise ValueError(
                    "Pivot transformer is only applicable to sequences, "
                    f"however, the value of `{str(match.full_path)}` "
                    "includes a non-sequence element."
                )

    def _transformer(item: TDataItem) -> TDataItem:
        """Pivot the given sequence item into a sequence of dicts.

        Args:
            item (TDataItem): The data item to transform.

        Returns:
            TDataItem: the data item with pivoted columns.
        """
        for path in paths:
            expr = jsonpath_ng.parse(path)

            for match in expr.find([item] if path in "$" else item):
                trans_value = []
                _raise_if_not_sequence(match)

                for value in match.value:
                    trans_value.append(_seq_to_dict(value))

                if path == "$":
                    item = trans_value
                else:
                    match.full_path.update(item, trans_value)

        return item

    return _transformer


def add_row_hash_to_table(row_hash_column_name: str) -> TDataItem:
    """Computes content hash for each row of pandas DataFrame, arrow table or batch and adds it as `row_hash_column_name` column.

    Internally arrow tables and batches are converted to pandas DataFrame and then `hash_pandas_object` is used to
    generate a series with row hashes. Hashes are converted to signed int64 and added to original table. Data may be modified.
    For SCD2 use with a resource configuration that assigns custom row version column to `row_hash_column_name`
    """
    from dlt.common.libs import pyarrow
    from dlt.common.libs.pyarrow import pyarrow as pa
    from dlt.common.libs.pandas import pandas as pd

    def _unwrap(table: TDataItem) -> TDataItem:
        if is_arrow := pyarrow.is_arrow_item(table):
            df = table.to_pandas(deduplicate_objects=False)
        else:
            df = table

        hash_ = pd.util.hash_pandas_object(df, index=False)

        if is_arrow:
            table = pyarrow.append_column(
                table,
                row_hash_column_name,
                pa.Array.from_pandas(hash_, type=pa.int64(), safe=False),
            )
        else:
            hash_np = hash_.values.astype("int64", copy=False, casting="unsafe")
            table[row_hash_column_name] = hash_np

        return table

    return _unwrap



from typing import Any, Dict, Optional, Type, Union, TYPE_CHECKING

from dlt.common.destination.configuration import ParquetFormatConfiguration
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.destination.capabilities import DataTypeMapper
from dlt.common.exceptions import MissingDependencyException
from dlt.destinations.impl.lancedb.configuration import (
    LanceDBCredentials,
    LanceDBClientConfiguration,
    TEmbeddingProvider,
)

LanceDBTypeMapper: Type[DataTypeMapper]
try:
    # lancedb type mapper cannot be used without pyarrow installed
    from dlt.destinations.impl.lancedb.type_mapper import LanceDBTypeMapper
except MissingDependencyException:
    # assign mock type mapper if no arrow
    from dlt.common.destination.capabilities import UnsupportedTypeMapper as LanceDBTypeMapper


if TYPE_CHECKING:
    from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient
    from lancedb import DBConnection


class lancedb(Destination[LanceDBClientConfiguration, "LanceDBClient"]):
    spec = LanceDBClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "parquet"
        caps.supported_loader_file_formats = ["parquet", "reference"]
        caps.type_mapper = LanceDBTypeMapper

        caps.max_identifier_length = 200
        caps.max_column_identifier_length = 1024
        caps.max_query_length = 8 * 1024 * 1024
        caps.is_max_query_length_in_bytes = False
        caps.max_text_data_type_length = 8 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = False
        caps.supports_ddl_transactions = False

        caps.decimal_precision = (38, 18)
        caps.timestamp_precision = 6
        caps.supported_replace_strategies = ["truncate-and-insert"]

        caps.recommended_file_size = 128_000_000

        caps.supported_merge_strategies = ["upsert"]

        # enable creation of nested types to support own vectors
        caps.supports_nested_types = True

        # must store arrow-compatible nested types, not parquet default - otherwise schema checker in lance fails
        caps.parquet_format = ParquetFormatConfiguration(use_compliant_nested_type=False)

        return caps

    @property
    def client_class(self) -> Type["LanceDBClient"]:
        from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient

        return LanceDBClient

    def __init__(
        self,
        credentials: Union["DBConnection", LanceDBCredentials, Dict[str, Any]] = None,
        lance_uri: Optional[str] = None,
        embedding_model_provider: TEmbeddingProvider = None,
        embedding_model: str = None,
        vector_field_name: str = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the LanceDB destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union["DBConnection", LanceDBCredentials, Dict[str, Any]]): Credentials to connect to the LanceDB database. Can be
                an instance of `LanceDBCredentials` or
                an instance of native LanceDB client or
                a dictionary with the credentials parameters.
            lance_uri (Optional[str]): LanceDB database URI. Defaults to local, on-disk instance.
                The available schemas are:
                - `/path/to/database` - local database.
                - `db://host:port` - remote database (LanceDB cloud).
            embedding_model_provider (TEmbeddingProvider, optional): Embedding provider used for generating embeddings.
                Default is "cohere". See LanceDB documentation for the full list of available providers.
            embedding_model (str, optional): The model used by the embedding provider for generating embeddings.
                Default is "embed-english-v3.0". Check with the embedding provider which options are available.
            vector_field_name (str, optional): Name of the special field to store the vector embeddings.
                Default is "vector".
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any, optional): Additional arguments forwarded to the destination config
        """
        super().__init__(
            credentials=credentials,
            lance_uri=lance_uri,
            embedding_model_provider=embedding_model_provider,
            embedding_model=embedding_model,
            vector_field_name=vector_field_name,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


lancedb.register()



import gzip
import contextlib
from typing import ClassVar, Iterator, List, IO, Any, Optional, Type, Generic

from dlt.common.metrics import DataWriterMetrics
from dlt.common.typing import TDataItem, TDataItems
from dlt.common.data_writers.exceptions import (
    BufferedDataWriterClosed,
    DestinationCapabilitiesRequired,
    FileImportNotFound,
    InvalidFileNameTemplateException,
)
from dlt.common.data_writers.writers import TWriter, DataWriter, FileWriterSpec, count_rows_in_items
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.configuration import with_config, known_sections, configspec
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.time import MonotonicPreciseTime
from dlt.common.utils import uniq_id


def new_file_id() -> str:
    """Creates new file id which is globally unique within table_name scope"""
    return uniq_id(5)


class BufferedDataWriter(Generic[TWriter]):
    @configspec
    class BufferedDataWriterConfiguration(BaseConfiguration):
        buffer_max_items: int = 5000
        file_max_items: Optional[int] = None
        file_max_bytes: Optional[int] = None
        disable_compression: bool = False
        _caps: Optional[DestinationCapabilitiesContext] = None

        __section__: ClassVar[str] = known_sections.DATA_WRITER

    @with_config(spec=BufferedDataWriterConfiguration)
    def __init__(
        self,
        writer_spec: FileWriterSpec,
        file_name_template: str,
        *,
        buffer_max_items: int = 5000,
        file_max_items: int = None,
        file_max_bytes: int = None,
        disable_compression: bool = False,
        _caps: DestinationCapabilitiesContext = None,
    ):
        self.writer_spec = writer_spec
        if self.writer_spec.requires_destination_capabilities and not _caps:
            raise DestinationCapabilitiesRequired(self.writer_spec.file_format)
        self.writer_cls = DataWriter.writer_class_from_spec(writer_spec)
        self._supports_schema_changes = self.writer_spec.supports_schema_changes
        self._caps = _caps
        # validate if template has correct placeholders
        self.file_name_template = file_name_template
        self.closed_files: List[DataWriterMetrics] = []  # all fully processed files
        # buffered items must be less than max items in file
        self.buffer_max_items = min(buffer_max_items, file_max_items or buffer_max_items)
        # Explicitly configured max size supersedes destination limit
        self.file_max_bytes = file_max_bytes
        if self.file_max_bytes is None and _caps:
            self.file_max_bytes = _caps.recommended_file_size
        self.file_max_items = file_max_items
        self.should_compress = self.writer_spec.supports_compression and not disable_compression
        # the open function is either gzip.open or open
        self.open = gzip.open if self.should_compress else open

        self._clock = MonotonicPreciseTime()
        self._current_columns: TTableSchemaColumns = None
        self._file_name: str = None
        self._buffered_items: List[TDataItem] = []
        self._buffered_items_count: int = 0
        self._writer: TWriter = None
        self._file: IO[Any] = None
        self._created: float = None
        self._last_modified: float = None
        self._closed = False
        try:
            self._rotate_file()
        except TypeError:
            raise InvalidFileNameTemplateException(file_name_template)

    def write_data_item(self, item: TDataItems, columns: TTableSchemaColumns) -> int:
        if self._closed:
            self._rotate_file()
            self._closed = False
        # rotate file if columns changed and writer does not allow for that
        # as the only allowed change is to add new column (no updates/deletes), we detect the change by comparing lengths
        if (
            self._current_columns is not None
            and (self._writer or self._supports_schema_changes == "False")
            and self._supports_schema_changes != "True"
            and len(columns) != len(self._current_columns)
        ):
            assert len(columns) > len(self._current_columns)
            self._rotate_file()
        # until the first chunk is written we can change the columns schema freely
        if columns is not None:
            self._current_columns = dict(columns)
        # add item to buffer and count new rows
        new_rows_count = count_rows_in_items(item)
        if isinstance(item, list):
            # items coming in a single list will be written together, no matter how many there are
            self._buffered_items.extend(item)
        else:
            self._buffered_items.append(item)
        self._buffered_items_count += new_rows_count
        # set last modification date
        self._last_modified = self._clock()
        # flush if max buffer exceeded, the second path of the expression prevents empty data frames to pile up in the buffer
        if (
            self._buffered_items_count >= self.buffer_max_items
            or len(self._buffered_items) >= self.buffer_max_items
        ):
            self._flush_items()
        # rotate the file if max_bytes exceeded
        if self._file:
            # rotate on max file size
            if self.file_max_bytes and self._file.tell() >= self.file_max_bytes:
                self._rotate_file()
            # rotate on max items
            elif self.file_max_items and self._writer.items_count >= self.file_max_items:
                self._rotate_file()
        return new_rows_count

    def write_empty_file(self, columns: TTableSchemaColumns) -> DataWriterMetrics:
        """Writes empty file: only header and footer without actual items. Closed the
        empty file and returns metrics. Mind that header and footer will be written."""
        self._rotate_file()
        if columns is not None:
            self._current_columns = dict(columns)
        self._last_modified = self._clock()
        return self._rotate_file(allow_empty_file=True)

    def import_file(
        self, file_path: str, metrics: DataWriterMetrics, with_extension: str = None
    ) -> DataWriterMetrics:
        """Import a file from `file_path` into items storage under a new file name. Does not check
        the imported file format. Uses counts from `metrics` as a base. Logically closes the imported file

        The preferred import method is a hard link to avoid copying the data. If current filesystem does not
        support it, a regular copy is used.

        Alternative extension may be provided via `with_extension` so various file formats may be imported into the same folder.
        """
        # TODO: we should separate file storage from other storages. this creates circular deps
        from dlt.common.storages import FileStorage

        # import file with alternative extension
        spec = self.writer_spec
        if with_extension:
            spec = self.writer_spec._replace(file_extension=with_extension)
        with self.alternative_spec(spec):
            (
                self._rotate_file(mark_compressed=True)
                if FileStorage.is_gzipped(file_path)
                else self._rotate_file(mark_compressed=False)
            )
        try:
            FileStorage.link_hard_with_fallback(file_path, self._file_name)
        except FileNotFoundError as f_ex:
            raise FileImportNotFound(file_path, self._file_name) from f_ex

        self._last_modified = self._clock()
        metrics = metrics._replace(
            file_path=self._file_name,
            created=self._created,
            last_modified=self._last_modified or self._created,
        )
        self.closed_files.append(metrics)
        # reset current file
        self._file_name = None
        self._last_modified = None
        self._created = None
        # get ready for a next one
        self._rotate_file()
        return metrics

    def close(self, skip_flush: bool = False) -> None:
        """Flushes the data, writes footer (skip_flush is True), collects metrics and closes the underlying file."""
        # like regular files, we do not except on double close
        if not self._closed:
            self._flush_and_close_file(skip_flush=skip_flush)
            self._closed = True

    @property
    def closed(self) -> bool:
        return self._closed

    @contextlib.contextmanager
    def alternative_spec(self, spec: FileWriterSpec) -> Iterator[FileWriterSpec]:
        """Temporarily changes the writer spec ie. for the moment file is rotated"""
        old_spec = self.writer_spec
        try:
            self.writer_spec = spec
            yield spec
        finally:
            self.writer_spec = old_spec

    def __enter__(self) -> "BufferedDataWriter[TWriter]":
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: Any) -> None:
        # skip flush if we had exception
        in_exception = exc_val is not None
        try:
            self.close(skip_flush=in_exception)
        except Exception:
            if not in_exception:
                # close again but without flush
                self.close(skip_flush=True)
            # raise the on close exception if we are not handling another exception
            if not in_exception:
                raise

    def _rotate_file(
        self,
        allow_empty_file: bool = False,
        mark_compressed: Optional[bool] = None,
    ) -> DataWriterMetrics:
        metrics = self._flush_and_close_file(allow_empty_file)

        base_filename = self.file_name_template % new_file_id()
        file_extension = self.writer_spec.file_extension

        # Add .gz extension if marked as a compressed imported file or
        # or no imported file mark is provided and configs instruct to compress
        if mark_compressed or (mark_compressed is None and self.should_compress):
            self._file_name = f"{base_filename}.{file_extension}.gz"
        else:
            self._file_name = f"{base_filename}.{file_extension}"

        self._created = self._clock()
        return metrics

    def _flush_items(self, allow_empty_file: bool = False) -> None:
        if self._buffered_items or allow_empty_file:
            # we only open a writer when there are any items in the buffer and first flush is requested
            if not self._writer:
                # create new writer and write header
                if self.writer_spec.is_binary_format:
                    self._file = self.open(self._file_name, "wb")  # type: ignore
                else:
                    self._file = self.open(self._file_name, "wt", encoding="utf-8", newline="")
                self._writer = self.writer_cls(self._file, caps=self._caps)  # type: ignore[assignment]
                self._writer.write_header(self._current_columns)
            # swap out buffer before writing so batch references are released
            # as soon as write_data returns, without waiting for the next
            # write_data_item call.
            if self._buffered_items:
                items = self._buffered_items
                self._buffered_items = []
                self._buffered_items_count = 0
                self._writer.write_data(items)
                items.clear()
            else:
                self._buffered_items_count = 0

    def _flush_and_close_file(
        self, allow_empty_file: bool = False, skip_flush: bool = False
    ) -> DataWriterMetrics:
        if not skip_flush:
            # if any buffered items exist, flush them
            self._flush_items(allow_empty_file)
            # if writer exists then close it
            if not self._writer:
                return None
            # write the footer of a file
            self._writer.write_footer()
            self._file.flush()
        else:
            if not self._writer:
                return None
        self._writer.close()
        # add file written to the list so we can commit all the files later
        metrics = DataWriterMetrics(
            self._file_name,
            self._writer.items_count,
            self._file.tell(),
            self._created,
            self._last_modified,
        )
        self.closed_files.append(metrics)
        self._file.close()
        self._writer = None
        self._file = None
        self._file_name = None
        self._created = None
        self._last_modified = None
        return metrics

    def _ensure_open(self) -> None:
        if self._closed:
            raise BufferedDataWriterClosed(self._file_name)
